from collections.abc import Sequence

from ice.formatter.multi import format_multi
from ice.formatter.multi import stop
from ice.formatter.transform.dependent import plural_transform
from ice.formatter.transform.positional import OrdinalWord
from ice.formatter.transform.value import numbered_list
from ice.recipes.experiments_and_arms.prompts.utils import get_part
from ice.recipes.experiments_and_arms.types import Arm
from ice.recipes.experiments_and_arms.types import Experiment
from ice.recipes.experiments_and_arms.types import ExperimentsArms
from ice.recipes.experiments_and_arms.types import Sample

SCHRIER: dict[str, ExperimentsArms] = {
    "gold_standard": ExperimentsArms(
        experiments=[
            Experiment(
                name="SALT-1",
                description="multicenter, randomized, double-blind, placebo-controlled trial of tolvaptan for hyponatremia",
                arms=[
                    Arm(
                        name="tolvaptan",
                        description="15mg tablet of oral tolvaptan daily, dose increased if necessary",
                        sample=Sample(size=102, stage="randomized"),
                    ),
                    Arm(
                        name="placebo",
                        description="tablet matching the tolvaptan tablet",
                        sample=Sample(size=103, stage="randomized"),
                    ),
                ],
            ),
            Experiment(
                name="SALT-2",
                description="multicenter, randomized, double-blind, placebo-controlled trial of tolvaptan for hyponatremia",
                arms=[
                    Arm(
                        name="tolvaptan",
                        description="15mg tablet of oral tolvaptan daily, dose increased if necessary",
                        sample=Sample(size=123, stage="randomized"),
                    ),
                    Arm(
                        name="placebo",
                        description="tablet matching the tolvaptan tablet",
                        sample=Sample(size=120, stage="randomized"),
                    ),
                ],
            ),
        ]
    ),
    "computed": ExperimentsArms(
        experiments=[
            Experiment(
                name="a study of tolvaptan",
                description="",
                arms=[
                    Arm(name="Tolvaptan", description="Placebo", sample=None),
                    Arm(
                        name='The "Tolvaptan" trial arm was the group of patients who were given the tolvaptan drug as part of the study. The intervention for this group was that they were given the tolvaptan drug for 4 days and the dose could be increased according to a regimen designed for slow correction of serum sodium concentrations.',
                        description='The "Placebo" trial arm was set up to receive a matching placebo once daily for up to 30 days. The placebo was given in the morning in either an inpatient or outpatient setting as an adjunct to the patient\'s standard therapy. Fluid restriction was not mandatory according to the study protocol. Treatment of hyponatremia with demeclocycline, lithium chloride, or urea was not permitted.',
                        sample=None,
                    ),
                ],
            ),
            Experiment(
                name="a study of the effect of discontinuation of the study drug",
                description="",
                arms=[
                    Arm(
                        name="Tolvaptan group", description="Placebo group", sample=None
                    ),
                    Arm(
                        name="The Tolvaptan group received the study drug while the Placebo group did not. The study drug was given in increasing doses over 4 days until the patient's serum sodium concentration reached 135 mmol per liter or more. If the serum sodium concentration remained below 136 mmol per liter and had increased by less than 5 mmol per liter during the prior 24 hours, the dose was increased. If the serum sodium concentration rose above 145 mmol per liter or increased at too great a rate (by more than 12 mmol per liter during 24 hours or by more than 8 mmol per liter during 8 hours on the first day of therapy), the investigator either withheld or decreased the next dose or increased the patient's fluid intake.",
                        description="The Placebo group was given a placebo for the first 30 days and then had the study drug withheld on day 30. The effect of discontinuation of the study drug was assessed on day 37.",
                        sample=None,
                    ),
                ],
            ),
        ]
    ),
}

HAUGEN: dict[str, ExperimentsArms] = {
    "gold_standard": ExperimentsArms(
        experiments=[
            Experiment(
                name="Surgical Safety Checklist (SSC) experiment",
                description="a stepped wedge cluster RCT to examine the effect of the SSC on in-hospital complications",
                arms=[
                    Arm(
                        name="preintervention cases",
                        description="surgical cases from before the checklist was implemented for a surgical specialty",
                        sample=Sample(size=2212, stage="randomized"),
                    ),
                    Arm(
                        name="postintervention cases",
                        description="surgical cases from after the checklist was implemented for a surgical specialty",
                        sample=Sample(size=3083, stage="randomized"),
                    ),
                ],
            )
        ]
    ),
    "computed": ExperimentsArms(
        experiments=[
            Experiment(
                name="a study of the WHO SSC",
                description="",
                arms=[
                    Arm(
                        name="Cardiothoracic surgery",
                        description="Neurosurgery",
                        sample=None,
                    ),
                    Arm(
                        name='The Cardiothoracic surgery trial arm was one of 5 trial arms in the experiment described as a study of the WHO SSC. The other trial arms were Neurosurgery, Orthopedic surgery, General surgery, and Urologic surgery. The intervention in the Cardiothoracic surgery trial arm was the implementation of the WHO Surgical Safety Checklist (SSC). The SSC consisted of 20 items and was performed at 3 critical steps of the surgical procedure: the "sign in" before induction of anesthesia, the "time out" before start of surgery, and the "sign out" before the head surgeon left the operating room. The SSC was introduced in all specialties/hospitals while using an educational program with standardized lectures and information materials.',
                        description='The "Neurosurgery" trial arm in the experiment described as a study of the WHO SSC was a cluster randomized controlled trial of the SSC. The intervention was sequentially rolled out in a random order until all 5 clusters-cardiothoracic, neurosurgery, orthopedic, general, and urologic surgery had received the Checklist. Data were prospectively recorded in control and intervention stages during a 10-month period in 2009-2010.',
                        sample=None,
                    ),
                ],
            )
        ]
    ),
}


def to_exps(result: ExperimentsArms) -> Sequence[str]:
    ret_val: list[str] = []
    for exp in result.experiments:
        if exp.description:
            ret_val.append(f"{exp.name}: {exp.description}")
        else:
            ret_val.append(exp.name)
    return ret_val


def display_result(result: ExperimentsArms, as_gs: bool) -> str:
    lines: list[str] = []
    for exp in result.experiments:
        lines.append("Experiment:")
        lines.append(f"{exp.name}: {exp.description}" if as_gs else exp.name)
        for arm in exp.arms:
            lines.append("Arms for this experiment:")
            lines.append(f"\t{arm.name}: {exp.description}" if as_gs else arm.name)
        lines.append("\n")
    return "\n".join(lines).strip()


EXAMPLES = [
    dict(
        gs_exps=numbered_list(to_exps(SCHRIER["gold_standard"])),
        gen_exps=numbered_list(to_exps(SCHRIER["computed"])),
        gs_exps_and_arms=display_result(SCHRIER["gold_standard"], as_gs=True),
        gen_exps_and_arms=display_result(SCHRIER["computed"], as_gs=False),
        differences="The generated answer and the gold standard both have two experiments, so they agree on the number. The model identified that one of these experiments studied tolvaptan, which is correct, but it identified the other experiment as studying 'discontinuation' of the drug, which does not appear on the gold standard. The model did correctly identify that each of the experiments had a placebo arm and a treatment arm.",
        grade="C",
    ),
    dict(
        gs_exps=numbered_list(to_exps(HAUGEN["gold_standard"])),
        gen_exps=numbered_list(to_exps(HAUGEN["computed"])),
        gs_exps_and_arms=display_result(HAUGEN["gold_standard"], as_gs=True),
        gen_exps_and_arms=display_result(HAUGEN["computed"], as_gs=False),
        differences="The generated answer and the gold standard both have one experiment. While the gold standard spells out the abbrevation, the generated answer does not. This is an extremely minor difference, so we should ignore it. The model's descriptions of the trial arms for this study, however, seem substantially incorrect: the gold standard identifies the arms as a postintervention and a preintervention group, while the model's outputs mention cardiothoracic surgery and then a long discussion, which does not even appear to be a specific arm (this output seems like the wrong shape).",
        grade="F",
    ),
]

SHARED_PROMPT_PARTS = dict(
    ordinal_word=OrdinalWord(capitalize=True),
    maybe_plural_gs_exp=plural_transform("gs_exps", "", "s"),
    maybe_plural_gen_exp=plural_transform("gen_exps", "", "s"),
)

EXPERIMENTS_EXAMPLE = """{ordinal_word}, consider the following gold standard, which describes the experiment{maybe_plural_gs_exp} conducted in a study:

{gs_exps_and_arms}

The machine learning model generated the following experiment{maybe_plural_gen_exp}:

{gen_exps_and_arms}

Explain the major differences. Ignore differences in formatting, or uses of abbreviations, or other minor differences. The model's outputs will generally not be as detailed as the gold standards, but you should not reduce scores for this unless the model's outputs are actually incorrect.

{differences}

On an A-F scale, where A is the best, and F is the worst, the model scored: {grade}"""

PREAMBLE = """Let's compare some of the generations from the machine learning model to our labeled gold standards. We're interested in knowing how well the model's results approximate the gold standards, so we'll ignore minor differences. The gold standards will include additional descriptions, which we are not using the model to generate. These additional descriptions are here only to provide context to help us better grade the model's outputs. Use the following grading scale:

A: Substantially correct in all respects
B: Agree on the number of outputs and the main thrust of their meanings, but with small differences.
C: Agree on the number of outputs, but disagree on some of them, while agreeing on others.
D: Disagree on the number of outputs or on the nature of all of them, but with similar themes.
F: Wild disagreement; not remotely close."""


def make_quick_eval_prompt(
    gs: ExperimentsArms, result: ExperimentsArms
) -> tuple[str, tuple[str, ...]]:
    last_example = dict(
        gs_exps=numbered_list(to_exps(gs)),
        gen_exps=numbered_list(to_exps(result)),
        gs_exps_and_arms=display_result(gs, as_gs=True),
        gen_exps_and_arms=display_result(result, as_gs=False),
        differences=stop(""),
    )
    parts = [PREAMBLE] + list(
        format_multi(
            EXPERIMENTS_EXAMPLE, EXAMPLES + [last_example], SHARED_PROMPT_PARTS  # type: ignore[arg-type]
        )
    )
    stop_seq = OrdinalWord(capitalize=True).transform(len(EXAMPLES), len(EXAMPLES) + 1)
    return "\n\n".join(parts), ("\n" + stop_seq,)


def get_grade(response: str) -> str:
    return get_part(response, "the model scored:", "NOTHING_LEFT")



from typing import Optional

from pydantic import BaseModel
from structlog.stdlib import get_logger

from ice.recipe import Recipe
from ice.utils import max_by_value

log = get_logger()


def make_compare_results_prompt(
    *, question: str, model_result: str, gold_result: str
) -> str:
    return f"""Compare the following results from a machine learning model to a gold standard. What (if any) information is in the gold standard but not in the model result? Does the model result capture all important information from the gold standard? Yes or No.

###

Question A: What was the placebo in this study?

Gold standard result A: Placebo gelatin capsules made of starch, talcum, kaolin, sucrose, and colouring. They were made by the same manufacturer as the active capsules and could not be distinguished from the active capsules by sight

Model result A: The placebo was made by the same manufacturer as the active capsules.

What (if any) information is in gold standard A but not in model result A?
- The specific composition of the placebo gelatin capsules (starch, talcum, kaolin, sucrose, and colouring)
- The fact that the placebo capsules could not be distinguished from the active capsules by sight

###

Question B: What was the image processing method?

Gold standard result B: The images were cleaned and aligned using common methods. The brain activity was measured using a statistical model that compared different types of stimuli (faces, houses, chairs, words, numbers, and a blank screen).

Model result B: The images were preprocessed using a standard pipeline that included skull stripping, motion correction, spatial normalization, and smoothing. The functional data were analyzed using a general linear model (GLM) with six task conditions (face, house, chair, word, number, and fixation) as regressors of interest and six motion parameters as nuisance regressors.

What (if any) information is in gold standard B but not in model result B? None

###

Question C: What was the test battery?

Gold standard result C: The participants completed a battery of neuropsychological tests that assessed memory, attention, executive function, language, and visuospatial skills. The scores were adjusted for age and education. They were expressed as standard deviations from the average.

Model result C: The participants took a series of tests that measured different aspects of their cognitive abilities, such as remembering words, repeating numbers, switching tasks, naming objects, and drawing a clock.

What (if any) information is in gold standard C but not in model result C?
- The analysis of test scores (adjusted for age and education, expressed as standard deviations from the average)

###

Question D: What was the study design?

Gold standard result D: The study used a randomized controlled trial design with four groups: a group that learned mindfulness techniques, a group that learned cognitive strategies, a group that learned both, and a group that received no treatment. The interventions lasted for eight weeks and consisted of meditation, yoga, cognitive exercises, and coping skills.

Model result D: The participants were randomly assigned to one of four groups: a mindfulness-based stress reduction (MBSR) group, a cognitive-behavioral therapy (CBT) group, a combination of MBSR and CBT (MBSR+CBT) group, or a wait-list control group. The MBSR group received eight weekly sessions of mindfulness meditation and yoga, the CBT group received eight weekly sessions of cognitive restructuring and coping skills training, and the MBSR+CBT group received both interventions.

What (if any) information is in gold standard D but not in model result D? None

###

Question E: {question.strip()}

Gold standard result E: {gold_result.strip()}

Model result E: {model_result.strip()}

What (if any) information is in gold standard E but not in model result E?"""


def make_classificaiton_prompt(
    *, question: str, missing_info: str, model_result: str, gold_result: str
) -> str:
    return f"""Compare the following results from a machine learning model to a gold standard. What (if any) information is in the gold standard but not in the model result? Does the model result capture all important information from the gold standard? Yes or No.

###

Question A: What was the placebo in this study?

Gold standard result A: Placebo gelatin capsules made of starch, talcum, kaolin, sucrose, and colouring. They were made by the same manufacturer as the active capsules and could not be distinguished from the active capsules by sight

Model result A: The placebo was made by the same manufacturer as the active capsules.

What (if any) information is in gold standard A but not in model result A?
- The specific composition of the placebo gelatin capsules (starch, talcum, kaolin, sucrose, and colouring)
- The fact that the placebo capsules could not be distinguished from the active capsules by sight

Does model result A capture all important information from gold standard A? No

###

Question B: What was the image processing method?

Gold standard result B: The images were cleaned and aligned using common methods. The brain activity was measured using a statistical model that compared different types of stimuli (faces, houses, chairs, words, numbers, and a blank screen).

Model result B: The images were preprocessed using a standard pipeline that included skull stripping, motion correction, spatial normalization, and smoothing. The functional data were analyzed using a general linear model (GLM) with six task conditions (face, house, chair, word, number, and fixation) as regressors of interest and six motion parameters as nuisance regressors.

What (if any) information is in gold standard B but not in model result B? None

Does model result B capture all important information from gold standard B? Yes

###

Question C: What was the test battery?

Gold standard result C: The participants completed a battery of neuropsychological tests that assessed memory, attention, executive function, language, and visuospatial skills. The scores were adjusted for age and education. They were expressed as standard deviations from the average.

Model result C: The participants took a series of tests that measured different aspects of their cognitive abilities, such as remembering words, repeating numbers, switching tasks, naming objects, and drawing a clock.

What (if any) information is in gold standard C but not in model result C?
- The analysis of test scores (adjusted for age and education, expressed as standard deviations from the average)

Does model result C capture all important information from gold standard C? No

###

Question D: What was the study design?

Gold standard result D: The study used a randomized controlled trial design with four groups: a group that learned mindfulness techniques, a group that learned cognitive strategies, a group that learned both, and a group that received no treatment. The interventions lasted for eight weeks and consisted of meditation, yoga, cognitive exercises, and coping skills.

Model result D: The participants were randomly assigned to one of four groups: a mindfulness-based stress reduction (MBSR) group, a cognitive-behavioral therapy (CBT) group, a combination of MBSR and CBT (MBSR+CBT) group, or a wait-list control group. The MBSR group received eight weekly sessions of mindfulness meditation and yoga, the CBT group received eight weekly sessions of cognitive restructuring and coping skills training, and the MBSR+CBT group received both interventions.

What (if any) information is in gold standard D but not in model result D? None

Does model result D capture all important information from gold standard D? Yes

###

Question E: {question}

Gold standard result E: {gold_result.strip()}

Model result E: {model_result.strip()}

What (if any) information is in gold standard E but not in model result E? {missing_info}

Does model result E capture all important information from gold standard E?"""


class ResultComparison(BaseModel):
    missing_info: Optional[str] = None
    is_complete: bool
    p_complete: float


class EvaluateResult(Recipe):
    async def run(
        self,
        model_result: Optional[str] = None,
        gold_result: Optional[str] = None,
        question: Optional[str] = None,
    ) -> ResultComparison:

        if self.mode == "test":
            model_results, gold_results, question = self.test_data()
            model_result = model_results[0]
            gold_result = gold_results[0]
        else:
            assert model_result is not None
            assert gold_result is not None
            assert question is not None

        missing_info_prompt = make_compare_results_prompt(
            question=question, model_result=model_result, gold_result=gold_result
        )
        missing_info = await self.agent().complete(
            prompt=missing_info_prompt,
            max_tokens=200,
        )
        classification_prompt = make_classificaiton_prompt(
            question=question,
            missing_info=missing_info,
            model_result=model_result,
            gold_result=gold_result,
        )
        choice_probs, _ = await self.agent("instruct").classify(
            prompt=classification_prompt, choices=(" Yes", " No")
        )

        choice, score = max_by_value(choice_probs)

        choice = choice.strip()

        if choice == "Yes":
            return ResultComparison(
                is_complete=True,
                p_complete=score,
                missing_info=missing_info,
            )
        elif choice == "No":
            return ResultComparison(
                is_complete=False,
                p_complete=1 - score,
                missing_info=missing_info,
            )
        else:
            raise ValueError(f"Unknown choice: {choice}")

    def test_data(self, n: int = 1) -> tuple[list[str], list[str], str]:
        if n > 3:
            raise ValueError("n must be <= 3")
        model_results = [
            "Placebo gelatin capsules made of starch, talcum, kaolin, sucrose, and colouring. They were made by the same manufacturer as the active capsules and could not be distinguished from the active capsules by sight",
            "The placebo consisted of chloroquine, proguanil, iron and folic acid tablets that had no active substances in them. They were identical in appearance and taste to the active tablets and were produced by the same company.",
            "The placebo was a fake treatment that looked like the real one but did not have any effect. It was made of chloroquine, proguanil, iron and folic acid tablets that had no active ingredients. The same company that made the real tablets also made the placebo tablets.",
        ]
        gold_results = [
            "The placebo is the treatment given to the control group that is designed to be indistinguishable from the real treatment but has no real effect. So in this case, the placebo for Group 5 would be chloroquine, proguanil, iron and folic acid pills that look identical to the real pills but contain no active ingredients. So the subjects would think they are taking the real treatment but are actually taking inactive pills.",
            "The placebo was chloroquine, proguanil, iron and folic acid tablets that had no active substances in them. It was identical in appearance and taste to the active tablets and were produced by the same company.",
            "Placebo gelatin capsules made of starch, talcum, kaolin, sucrose, and colouring. They were made by the same manufacturer as the active capsules and could not be distinguished from the active capsules by sight",
        ]
        question = "What was the placebo in the study?"
        return model_results[:n], gold_results[:n], question



from ice.recipe import recipe
from ice.recipes.primer.amplify_one.utils import *
from ice.recipes.primer.subquestions import ask_subquestions
from ice.utils import map_async


async def get_subs(question: str) -> Subs:
    subquestions = await ask_subquestions(question=question)
    subanswers = await map_async(subquestions, answer)
    return list(zip(subquestions, subanswers))


async def answer(question: str, subs: Subs = []) -> str:
    prompt = make_qa_prompt(question, subs=subs)
    answer = await recipe.agent().complete(prompt=prompt, stop='"')
    return answer


async def answer_by_amplification(
    question: str = "What is the effect of creatine on cognition?",
):
    subs = await get_subs(question)
    response = await answer(question=question, subs=subs)
    return response


recipe.main(answer_by_amplification)



import re

from dataclasses import dataclass
from typing import Literal

from ice.evaluation.evaluate_recipe_result import RecipeResult
from ice.metrics.gold_standards import list_experiments
from ice.paper import Paper
from ice.recipe import Recipe

PlaceboClassification = Literal["Placebo", "Not mentioned", "No placebo", "Unclear"]


@dataclass
class PlaceboAnswer:
    classification: PlaceboClassification
    sentence: str


class PlaceboKeywordBaseline(Recipe):
    @staticmethod
    def answer_placebo_using_keywords(paper: Paper) -> PlaceboAnswer:
        """
        1. first, if the paper says it's open-label (matches `open[- ](control|label)`), classify as "No placebo"
            a. return the 1st sentence that matches as the quote supporting this answer
        2. then, if we didn't classify as "No placebo" based on the previous step, then check if the world "placebo" is in the paper. If so, classify as "Placebo"
            a. return the 1st sentence that contains "placebo" as the quote supporting this answer as well as the description of the placebo/the answer
        3. then, if we haven't found anything yet, classify as "Not mentioned"
            a. don't return any supporting quote
        """
        sentences = [
            sentence
            for paragraph in paper.paragraphs
            for sentence in paragraph.sentences
        ]
        for sentence in sentences:
            if re.search("open[- ](control|label)", sentence, re.IGNORECASE):
                return PlaceboAnswer("No placebo", sentence)
        for sentence in sentences:
            if re.search("placebo", sentence, re.IGNORECASE):
                return PlaceboAnswer("Placebo", sentence)
        return PlaceboAnswer("Not mentioned", "")

    async def run(self, paper: Paper):
        experiments = list_experiments(document_id=paper.document_id)

        results: list[RecipeResult] = []
        for experiment in experiments:
            placebo_answer = self.answer_placebo_using_keywords(paper)
            results.append(
                RecipeResult(
                    experiment=experiment,
                    question_short_name="placebo",
                    document_id=paper.document_id,
                    result=f"{placebo_answer.classification}: {placebo_answer.sentence}",
                    answer=f"{placebo_answer.classification}: {placebo_answer.sentence}",
                    classifications=[
                        "Placebo"
                        if placebo_answer.classification == "Placebo"
                        else "No placebo or placebo not mentioned",
                        placebo_answer.classification,
                    ],
                    excerpts=[placebo_answer.sentence],
                )
            )

        self.maybe_add_to_results(results)

        return results



from dataclasses import dataclass
from typing import Any
from typing import cast
from typing import Optional

from ice.json_value import JSONValue


@dataclass
class Summarizer:
    str_limit: int = 100
    list_limit: int = 3
    dict_limit: int = 10
    depth_limit: int = 4
    float_digits: int = 4

    def summarize(self, x: JSONValue, depth_left: Optional[int] = None):
        if depth_left is None:
            depth_left = self.depth_limit
        elif depth_left <= 0:
            return None
        if isinstance(x, list):
            return self.summarize_list(x, depth_left)
        elif isinstance(x, dict):
            return self.summarize_dict(x, depth_left)
        elif isinstance(x, str):
            return self.summarize_str(x)
        elif isinstance(x, float):
            return self.summarize_float(x)
        else:
            return x

    def summarize_dict(self, x: dict[str, JSONValue], depth_left: int):
        if list(x.keys()) == ["__fstring__"]:
            new_x = {}
            for part in cast(Any, x["__fstring__"]):
                if isinstance(part, dict) and "value" in part:
                    new_x[part["source"]] = part["value"]
            x = new_x

        result = {}
        for k in sorted(x.keys())[: self.dict_limit]:
            v = self.summarize(x[k], depth_left - 1)
            if not self._is_empty(v):
                result[k] = v
        return result

    def summarize_list(self, x: list[JSONValue], depth_left: int):
        result = [self.summarize(v, depth_left - 1) for v in x[: self.list_limit]]
        while result and self._is_empty(result[-1]):
            result.pop()
        return result

    def summarize_str(self, string: str):
        """
        Returns a version of `string` at most `self.str_limit` characters long,
        with the middle replaced by `...` if necessary.
        """
        lim = self.str_limit
        if len(string) > lim:
            # uncomment to omit long strings entirely instead of truncating
            # return None

            middle = "..."
            i = (lim - len(middle)) // 2
            j = lim - len(middle) - i
            string = string[:i] + middle + string[-j:]
        return string

    def summarize_float(self, x: float):
        # TODO use significant digits instead of decimal places?
        return round(x, self.float_digits)

    def _is_empty(self, x: JSONValue):
        return x in ([], {}, "", None)


summarize = Summarizer().summarize



from structlog import get_logger

from ice.recipe import Recipe

log = get_logger()


class MesaRecipe(Recipe):
    async def run(self, input: list[str]) -> list[str]:
        assert input, "MesaRecipe requires an 'input' argument"

        async def first_node():
            output = input + ["MesaRecipe.first_node"]
            return output

        async def second_node(pred: list[str]):
            output = pred + ["MesaRecipe.second_node"]
            return output

        return await second_node(pred=(await first_node()))


class ExampleMetaRecipe(Recipe):
    async def run(self):
        other_recipe = MesaRecipe(mode=self.mode)
        sub_result = await other_recipe.run(input=["MetaRecipe"])
        final_result = sub_result + ["MetaRecipe.collect_result"]
        return " ---> ".join(final_result)



from typing import Optional

from ice.agents.base import Agent
from ice.agents.base import Stop
from ice.environment import env
from ice.utils import max_by_value
from ice.utils import quoted


class AugmentedAgent(Agent):
    def __init__(self, human: Agent, machine: Agent):
        self.human = human
        self.machine = machine

    async def relevance(
        self, *, question, context, verbose=False, default=None
    ) -> float:
        machine_resp = await self.machine.relevance(
            question=question, context=context, verbose=verbose
        )
        human_resp = await self.human.relevance(
            question=question, context=context, verbose=verbose, default=machine_resp
        )
        return human_resp

    async def complete(
        self,
        *,
        prompt: str,
        stop: Stop = None,
        verbose: bool = False,
        default: str = "",
        max_tokens: int = 256,
    ):
        machine_resp = await self.machine.complete(
            prompt=prompt,
            stop=stop,
            verbose=verbose,
            default=default,
            max_tokens=max_tokens,
        )
        return await self.human.complete(
            prompt=prompt,
            stop=stop,
            verbose=verbose,
            default=machine_resp,
            max_tokens=max_tokens,
        )

    async def predict(self, *, context, default="", verbose=False) -> dict[str, float]:
        machine_resp: dict[str, float] = await self.machine.predict(
            context=context, verbose=verbose
        )
        # Extract most likely response from machine_resp and use it as default for human.
        most_likely_token = max(machine_resp, key=lambda k: machine_resp[k])
        human_resp = await self.human.predict(
            context=context, default=most_likely_token, verbose=verbose
        )
        return human_resp

    async def classify(
        self,
        *,
        prompt: str,
        choices: tuple[str, ...],
        default: Optional[str] = None,
        verbose: bool = False,
    ) -> tuple[dict[str, float], Optional[str]]:
        (machine_probs, explanation) = await self.machine.classify(
            prompt=prompt,
            choices=choices,
            default=default,
            verbose=verbose,
        )
        machine_choice, _ = max_by_value(machine_probs)
        if explanation is not None:
            # TODO: Should present this to the human in a way that does
            #       not circumvent agent abstraction
            env().print(
                f"""
#### classify

Machine choice probs:

{machine_probs}

Explanation for machine probs:

{quoted(explanation)}""",
                format_markdown=True,
            )
        return await self.human.classify(
            prompt=prompt,
            choices=choices,
            default=machine_choice,
            verbose=verbose,
        )



from typing import Optional
from urllib.parse import urljoin

from structlog.stdlib import get_logger

from ice.recipe import recipe
from ice.recipes.elicit.common import send_elicit_request
from ice.recipes.elicit.search import get_elicit_backend

log = get_logger()

ELICIT_QA_ENDPOINT = "full-text-qa"


def get_span_container(paper, source):
    return paper[source] if source == "abstract" else paper["body"]["value"]


def get_span_paragraph(container, paragraph_index):
    return container["paragraphs"][paragraph_index]


def get_span_text(paragraph, sentence_start, sentence_stop):
    sentences = paragraph["sentences"][sentence_start:sentence_stop]
    return " ".join(sentences)


def augment_qa_response(response, papers):
    for paper_id, paper_data in response["papers"].items():
        for qa_key, qa_data in paper_data["fullTextQaResults"].items():
            # Skip if the query failed or no excerpt was found
            if qa_data["status"] != "success" or not qa_data["value"].get(
                "mostRelevantExcerpt"
            ):
                continue
            # Get the span and its source, paragraph index, and sentence range
            span = qa_data["value"]["mostRelevantExcerpt"]["spans"][0]
            source, paragraph_index, sentence_start, sentence_stop = (
                span["source"],
                span["paragraphIndex"],
                span["sentenceStart"],
                span["sentenceStop"],
            )
            # Get the text of the span from the paper
            container = get_span_container(papers[paper_id], source)
            paragraph = get_span_paragraph(container, paragraph_index)
            span_text = get_span_text(paragraph, sentence_start, sentence_stop)
            # Add the span text to the response
            qa_data["value"]["mostRelevantExcerpt"]["text"] = span_text
    return response


async def elicit_qa(
    root_question: str = "What is the effect of creatine on cognition?",
    qa_question: str = "What was the population?",
    papers: Optional[dict] = None,
):
    cells = [
        dict(paper, column={"type": "custom_question", "value": qa_question})
        for paper in (papers or dict()).values()
    ]
    request_body = {"rootQuestion": root_question, "cells": cells}
    backend = await get_elicit_backend()
    endpoint = urljoin(backend, ELICIT_QA_ENDPOINT)
    response = send_elicit_request(request_body=request_body, endpoint=endpoint)
    # Augment the response with the text of the extract spans
    response = augment_qa_response(response, papers)
    return response


async def elicit_qa_cli(
    root_question: str = "What is the effect of creatine on cognition?",
    qa_question: str = "What was the population?",
):
    response = await elicit_qa(root_question, qa_question)
    return response


recipe.main(elicit_qa_cli)

# Example result:
# {
#     'papers': {
#         'deb4f9cf49a8f1d43e6db4c0cfe543c982911ded': {
#             'fullTextQaResults': {
#                 'custom_question-What was the key result?': {
#                     'status': 'success',
#                     'value': {
#                         'rephrasedQuestion': None,
#                         'mostRelevantExcerpt': {
#                             'spans': [
#                                 {
#                                     'source': 'abstract',
#                                     'paragraphIndex': 0,
#                                     'sentenceStart': 0,
#                                     'sentenceStop': None
#                                 }
#                             ],
#                             'text': 'Background and aims: Creatine is a supplement used by sportsmen to increase athletic performance by improving energy supply to muscle tissues. It is also an essential brain compound and some hypothesize that it aids cognition by improving energy supply and neuroprotection. The aim of this systematic review is to investigate the effects of oral creatine administration on cognitive function in healthy individuals. Methods: A search of multiple electronic databases was performed for the identification of randomized clinical trials (RCTs) examining the cognitive effects of oral creatine supplementation in healthy individuals. Results: Six studies (281 individuals) met our inclusion criteria. Generally, there was evidence that short term memory and intelligence/reasoning may be improved by creatine administration. Regarding other cognitive domains, such as long‐term memory, spatial memory, memory scanning, attention, executive function, response inhibition, word fluency, reaction time and mental fatigue, the results were conflicting. Performance on cognitive tasks stayed unchanged in young individuals. Vegetarians responded better than meat‐eaters in memory tasks but for other cognitive domains no differences were observed. Conclusions: Oral creatine administration may improve short‐term memory and intelligence/reasoning of healthy individuals but its effect on other cognitive domains remains unclear. Findings suggest potential benefit for aging and stressed individuals. Since creatine is safe, future studies should include larger sample sizes. It is imperative that creatine should be tested on patients with dementias or cognitive impairment. HIGHLIGHTSOral creatine supplementation improves memory of healthy adults.Findings suggest potential benefit for aging and stressed individuals.Future trials should investigate the effect of creatine administration on individuals with dementia or mild cognitive impairment.'
#                         },
#                         'answer': 'Oral creatine supplementation improves memory of healthy adults',
#                         'mostRelevantSpan': None,
#                         'subResults': None
#                     }
#                 }
#             }
#         }
#     }
# }



# I am an async thread-safe way to store high-level information about the
# current execution.
# I'm specific about my parameters because it would be too easy to hang all
# sorts of information off me, when all I'm supposed to do is know about the
# original parameters an execution was started with.
from contextvars import ContextVar
from typing import Optional
from typing import TypedDict

from ice.utils import make_id


class ExecutionContext(TypedDict):
    id: str
    document_id: Optional[str]
    task: str


__execution_context = ContextVar[ExecutionContext]("execution_context")


def new_context(*, document_id: Optional[str], task: str) -> None:
    context: ExecutionContext = {
        "id": make_id(),
        "document_id": document_id,
        "task": task,
    }
    global __execution_context
    __execution_context.set(context)


def context() -> ExecutionContext:
    return __execution_context.get()



from collections.abc import Callable
from collections.abc import Sequence
from typing import Generic
from typing import TypeVar
from typing import Union

from ice.formatter.transform import _Transform


T_contra = TypeVar("T_contra", contravariant=True)


class ValueTransform(_Transform, Generic[T_contra]):
    """
    Implement to functionally transform the value.
    """

    def __init__(
        self, value: T_contra, transform: Callable[[T_contra], Union[str, int]]
    ):
        self.value = value
        self._transform = transform

    def transform(self):
        return self._transform(self.value)


def non_literal(value: T_contra) -> ValueTransform[T_contra]:
    """
    Escape hatch to use a non-literal with `format_multi`
    """
    return ValueTransform(value, transform=lambda v: str(v))


def _paragraphs_to_numbered_list(paragraphs: Sequence[str], separator: str) -> str:
    return separator.join(
        f"{n}. {paragraph}".strip() for n, paragraph in enumerate(paragraphs, 1)
    )


def numbered_list(
    value: Sequence[str], separator: str = "\n"
) -> ValueTransform[Sequence[str]]:
    """Turn a sequence of strings into a single enumerated
    list, beginning with 1.
    """
    return ValueTransform(value, lambda v: _paragraphs_to_numbered_list(v, separator))



from collections.abc import Mapping
from collections.abc import Sized
from typing import Generic
from typing import Optional
from typing import TypeVar
from typing import Union

from numerizer.consts import DIRECT_SINGLE_NUMS

from ice.formatter.transform import _Transform


T_contra = TypeVar("T_contra", contravariant=True)


class DependentTransform(_Transform, Generic[T_contra]):
    """
    Implement to complete based on the value of another placeholder.
    """

    def key(self) -> str:
        raise NotImplementedError

    def transform(self, dependent: T_contra) -> str:
        raise NotImplementedError


class CountWord(DependentTransform):
    """
    Fill in based on the value of the placeholder for `key` if it is an integer,
    or else its size.

    By default these will be English counts ("one", "two", "three", etc.), up
    to "twenty". Override or add additional cases by passing in
    a `special` mapping.
    """

    def __init__(self, key: str, special: Optional[Mapping[int, str]] = None):
        self._key = key
        self.words = {int(v): k for k, v in DIRECT_SINGLE_NUMS.items()}
        if special:
            for count, word in special.items():
                self.words[count] = word

    def key(self):
        return self._key

    def transform(self, dependent: Union[int, Sized]) -> str:
        count = dependent if isinstance(dependent, int) else len(dependent)
        try:
            return self.words[count]
        except KeyError:
            raise ValueError(f"Count {count} not in count word dictionary")


def plural_transform(
    key: str, singular_case: str, plural_case: str
) -> DependentTransform[Union[int, Sized]]:
    """
    Return the singular or plural case based on the
    value (if an `int`) or size of the placeholder
    at `key`.
    """

    class SingularOrPlural(DependentTransform):
        def __init__(self, key: str):
            self._key = key

        def key(self):
            return self._key

        def transform(self, dependent: Union[int, Sized]) -> str:
            count = dependent if isinstance(dependent, int) else len(dependent)
            return singular_case if count == 1 else plural_case

    return SingularOrPlural(key)



from collections.abc import Sequence

from ice.recipe import recipe
from ice.recipes.experiments_and_arms.num_utils import strip_enumeration_prefix
from ice.recipes.experiments_and_arms.prompts.consensus import build_cluster_prompt
from ice.recipes.experiments_and_arms.prompts.consensus import build_final_prompt


async def best_answer_by_consensus(question: str, candidates: Sequence[str]) -> str:
    """Identify the best answer by first identifying clusters,
    then summarizing a consensus among the answers.

    Args:
        question (str): Question for the candidate answers.
        candidates (Sequence[str]): Candidate answers.

    Returns:
        str: The consensus answer according to the agent.
    """
    clusters = await recipe.agent().complete(
        prompt=build_cluster_prompt(question, candidates)
    )
    final_answer = await recipe.agent().complete(
        prompt=build_final_prompt(question, candidates, clusters)
    )

    return strip_enumeration_prefix(final_answer)


recipe.main(best_answer_by_consensus)



from collections.abc import Callable
from collections.abc import Sequence
from typing import Optional
from typing import Union

from ice.formatter.multi import format_multi
from ice.formatter.transform.dependent import CountWord
from ice.formatter.transform.dependent import plural_transform
from ice.formatter.transform.positional import OrdinalWord
from ice.formatter.transform.value import numbered_list
from ice.formatter.transform.value import ValueTransform
from ice.recipes.experiments_and_arms.prompts.utils import get_part
from ice.recipes.experiments_and_arms.prompts.utils import start_last_example
from ice.recipes.experiments_and_arms.types import MultipartReasoningPrompt


CAN_WE_NAME_ARMS_EXAMPLES: list[
    dict[str, Union[ValueTransform[Sequence[str]], str, int]]
] = [
    dict(
        paragraphs=numbered_list(
            [
                """As described above, prior research largely has focused on the types of threats to validity. Few explicit recommendations exist for quality control procedures (e.g., assertions of goldstandard approaches), and little is known about how specific quality assurance approaches affect study outcomes. Therefore, given the literature indicating the need to manage VPNs/bots, misrepresentation, and inattention, we conducted a randomized, controlled experiment with four study arms (Control Arm, Bot/VPN Arm, Truth/Attention Arm, and Stringent Arm). The purpose of the study was to determine the absolute numeric difference, as well as differences in magnitude, skewness, and standard deviation, of different quality control procedures on outcomes from three self-administered digital tools used in a cross section of fields focused on mental health and substance use: the U.S. Alcohol Use Disorder Identification Test (USAUDIT) (Higgins-Biddle & Babor, 2018), the Patient Health Questionnaire (PHQ-9) (Kroenke et al., 2001), and the screener for Generalized Anxiety Disorder (GAD-7) (Spitzer et al., 2006).""",
                """These instruments (or similar, such as the AUDIT-C) have been used in recent crowdsourced studies on MTurk to explore a variety of important topics, such as associations between loneliness, depression, and COVID-19 (Killgore et al., 2020), relationships between sleep debt and anxiety (Dickinson et al., 2018), temporal relationships between day-level cravings and alcohol use (Jain et al., 2021), and the efficacy of Internet interventions for unhealthy alcohol use (Cunningham et al., 2019). At the same time, clinical studies have noted differences in self-reported prevalence of depression and anxiety between adult MTurk samples and other data sources, such as adult community samples or undergraduate research samples. Often, but not always, anxiety and depression have appeared to be more prevalent in samples from MTurk, and researchers have encouraged exploration of why this might be the case (Arditte et al., 2016;van Stolk-Cooke et al., 2018;Ophir et al., 2019;Engle et al., 2020). In response to this need, our methodological research provides the rapidly growing number of scholars using crowdsourced sampling with objective data indicating the expected impact and utility of multiple different quality-control procedures on crowdsourced data assessing depression, anxiety, and risky alcohol use.""",
                """We proposed two exploratory, preregistered hypotheses .""",
                """(1) Outcome scores from each of the three screening tools would be significantly different for each pairwise comparison of study arms, except for the pairing (Bot/VPN with Truth/Attention). We expected that each additional form of quality control would affect outcomes on all screening tools, except that we were agnostic as to whether there would be a meaningful difference between the different types of quality control (e.g., that the Bot/VPN control and the Truth/Attention control would produce differential effects). Thus, our hypotheses were based on the stringency of control mechanics by frequency count (e.g., 0, 1, 1, or 2 approaches within the arm), and we expected differences between each pair except the 1:1 pairing.""",
            ]
        ),
        reasoning="\n\n".join(
            [
                "Excerpt 1 says that there were four arms and describes them: Control, Bot/VPN, Truth/Attention, and Stringent.",
                "Excerpt 2 provides background information but does not discuss arms.",
                "Excerpt 3 says that there were two preregistered hypotheses for the study",
                "Excerpt 4 discusses the first of these preregistered hypotheses and alludes to some of the aforementioned arms (Bot/VPN and Truth/Attention)",
            ]
        ),
        helpfulness="Excerpt 1 was extremely helpful; excerpts 2, 3, and 4 were not helpful.",
        experiments=numbered_list(
            [
                "An experiment to test whether quality control on MTurk affects outcomes relevant to psychological/behavioral research"
            ]
        ),
        experiment_in_question="An experiment to test whether quality control on MTurk affects outcomes relevant to psychological/behavioral research",
        answer="Yes",
    ),
    dict(
        paragraphs=numbered_list(
            [
                """To measure the effect of an electronic health record (EHR) alert on chronic hepatitis B (CHB) screening among at-risk Asian and Pacific Islanders (API). API patients who had not yet completed hepatitis B surface antigen (HBsAg) testing were identified by a novel EHR-based population health tool. At-risk API patients in Cohort 1 (primarily privately insured) and Cohort 2 (includes Medicare and/or Medicaid) were randomized to alert activation in their electronic medical charts or not. In total, 8299 API were found to be deficient in HBsAg completion at baseline within our health system. In Cohort 1, 1542 patients and 1568 patients were randomized to the alert and control respectively. In Cohort 2, 2599 patients and 2590 patients were randomized to the alert and control respectively. For both cohorts combined, 389 HBsAg tests were completed in the alert group compared to 177 HBsAg tests in the control group (p < 0.0001; OR = 2.3; 95% CI 1.94-2.80), but there was no increased detection of HBsAg positivity from the alert (15 versus 13 respectively, p = 0.09; OR = 0.5; 95% CI 0.24-1.09). Our results demonstrate that personalized, automated electronic alerts increase screening for CHB, but more comprehensive measures are needed to detect HBsAg positive patients. NIH Trial Registry Number: NCT04240678.""",
                """Based on the most recent analysis of National Health and Nutrition Examination Survey (NHANES), chronic hepatitis B (CHB) affects 847,000 persons in the United States, which included approximately 400,000 non-Hispanic Asians. These Asians had a tenfold greater prevalence of CHB than the American general population 1 . Foreign-born persons have the highest CHB prevalence in the United States, between 4.5 and 10.3%, and the majority of foreign-born persons with CHB living in the United States originated from Asia 2 . Since CHB is the leading cause of hepatocellular carcinoma and cirrhosis in the world 3 , it is not surprising that Asian and Pacific Islanders (API) living in the US have the highest rates of HCC and HCC-related death 4 .""",
                """Because of this, the Centers for Disease Control and Prevention (CDC), United States Preventive Services Task Force, and the American Association for Study of Liver Diseases have recommended screening all persons born in countries with CHB endemicity ≥ 2%. Despite these recommendations, screening rates for CHB remain low, which may be due in part to lack of physician awareness and knowledge about CHB guidelines [5][6][7][8] .""",
                """Because of this, the Centers for Disease Control and Prevention (CDC), United States Preventive Services Task Force, and the American Association for Study of Liver Diseases have recommended screening all persons born in countries with CHB endemicity ≥ 2%. Despite these recommendations, screening rates for CHB remain low, which may be due in part to lack of physician awareness and knowledge about CHB guidelines [5][6][7][8] .""",
            ]
        ),
        reasoning="\n\n".join(
            (
                "Excerpt 1 identifies the two arms for the Cohort 1 (and Cohort 2) study: patients were randomized to alert activation in their electronic medical charts (the 'alert' arm) or not (the control arm).",
                "Excerpt 2 provides background information",
                "Excerpt 3 profides additional background information",
                "Excerpt 4 describes the significance of the findings, alluding to the two groups mentioned above (alert and control), here without specifically mentioning the two studies.",
            )
        ),
        helpfulness="Excerpt 1 was extremely helpful, and excerpt 2, 3, and 4 were not helpful.",
        experiments=numbered_list(["Cohort 1", "Cohort 2"]),
        experiment_in_question="Cohort 1",
        answer="Yes",
    ),
    dict(
        paragraphs=numbered_list(
            [
                """Although the curricula were delivered in English, some children spoke primarily Spanish or Haitian-Creole, particularly at the younger assessment ages. Twenty-five children in Miami and 92 children in LA completed at least one assessment in Spanish, and some assessment items were translated into Haitian-Creole for 7 children in Miami. Children who were bilingual or monolingual non-English speaking were assessed by bilingual staff. Bilingual ability increased over time for foreign language speakers. By third grade, all child assessments were conducted in English. There was no racial or gender bias in the selection of participants.""",
                """Children's cognitive development was assessed directly using the following developmentally appropriate measures: Kaufman Assessment Battery for Children-II (KABC) 25 at ages 3 and 5 years and at third grade ( Y3, Y5, and third grade, respectively) and WoodcockJohnson (WJ) III Tests of Achievement 26 at Y5 and third grade. See Table 4 for a summary of cognitive and language measures and assessment time points.""",
                """All the cognitive composite scores are standardized to a mean of 100 and SD of 15. The KABC measures are widely used, 27 norm-referenced instruments designed to assess cognitive ability 25 with utility for bilingual children. The WJ Tests of Achievement are a comprehensive set of norm-referenced tests for measuring academic achievement. The WJ Tests of Achievement have good reliability and validity 28 and have been used in other large studies with low-income children, such as the Head Start Family and Child Experiences Survey. 29 The WJ Achievement subscales administered in this sample were Letter-Word Identification (Y5 and third grade), Spelling(y5)and Passage Comprehension (third grade) to measure reading skills and Applied Problems (Y5 and third grade) and Calculations (third grade) to measure mathematical skills.""",
                """Child language development was assessed directly with the Preschool Language Scale-4 (PLS) 30 at Y2 and Y4, with the Test of Early Reading Ability-3 (TERA) 31 at Y4, and through maternal report on the Adaptive Language Inventory (ALI) 32 at Y5. The PLS is a normreferenced instrument designed to assess expressive and receptive language skills; the PLS includes 2 subscales, Auditory Comprehension and Expressive Communication, as well as a Total Language score. The TERA is an assessment that measures young children's ability to attribute meaning to printed symbols, knowledge of the alphabet, and understanding of the conventions of print. The TERA has concurrent validity with other measures of reading achievement and verbal intelligence quotient (IQ) tests. 31 The PLS and TERA subscale scores are standardized to a mean of 100 and a SD of 15. The ALI is a rating scale to assess children's use of narrative and discourse skills. Mothers were interviewed on items that addressed the following domains: Comprehension, Expression, Rephrase, Listening, Spontaneity, and Fluency. Total scores for each of these domains range from 4 to 20 for Comprehension, Expression, and Rephrase and from 2 to 10 for Listening, Spontaneity, and Fluency. Alphas for the ALI scales ranged from 0.67 to 0.89 for this sample, indicating moderate to high internal consistency. Language was not separately assessed at the third grade follow-up assessment.""",
            ]
        ),
        reasoning="\n\n".join(
            (
                """Excerpt 1 discusses some findings but does not name trial arms, and it is not clear which experiment is being discussed.""",
                """Excerpt 2 discusses how results were measured but does not identify trial arms.""",
                """Excerpt 3 continues to discuss the assessment methodology""",
                """Excerpt 4 continues to discuss measurement methodology.""",
            )
        ),
        helpfulness="Excerpts 1, 2, 3, and 4 were not helpful.",
        experiments=numbered_list(
            [
                "an RCT of the Legacy for Children program in LA",
                "an RCT of the Legacy for Children program in Miami",
            ]
        ),
        experiment_in_question="an RCT of the Legacy for Children program in Miami",
        answer="No",
    ),
]

SHARED_CAN_WE_NAME_ARMS_PROMPT = dict(
    maybe_plural=plural_transform("experiments", "", "s"),
    order=OrdinalWord(capitalize=True),
    just_one_or_how_many=CountWord("experiments"),
    do_or_does_the_above_excerpts=plural_transform(
        "paragraphs", "does the above excerpt", "do the above excerpts"
    ),
    experiments_count=CountWord("experiments"),
    each_or_the=plural_transform("experiments", "the", "each"),
    were_or_was=plural_transform("experiments", "was", "were"),
    maybe_plural_excerpts=plural_transform("paragraphs", "", "s"),
)

CAN_WE_NAME_ARMS_TEMPLATE = """{order} consider the following excerpt{maybe_plural_excerpts} from a research paper that conducted {just_one_or_how_many} experiment{maybe_plural}:

{paragraphs}

In this paper, there {were_or_was} {experiments_count} experiment{maybe_plural}:

{experiments}

Do these excerpts identify the different trial arms (subgroups of participants) of the "{experiment_in_question}" experiment specifically?

Let's think it over:

{reasoning}

Which excerpts, if any, were helpful in identifying the trial arms (subgroups of participants) of the "{experiment_in_question}" experiment? {helpfulness}

Therefore, {do_or_does_the_above_excerpts} provide insight into what the trial arms were? f we cannot tell from the excerpt{maybe_plural_excerpts}, answer "Unclear". Answer Yes, No, or Unclear. {answer}
"""


def can_we_name_experiments_stop_seq(reasoning: Optional[str]) -> Sequence[str]:
    return ["\n\n"] if reasoning else ["Therefore, "]


def make_can_we_name_arms_prompt(
    experiments: Sequence[str],
    experiment_in_question: str,
) -> Callable[[int], MultipartReasoningPrompt]:
    def from_num_shot(num_shot: int):
        def make_can_we_name_arms_prompt(
            paragraphs: Sequence[str],
            helpfulness: Optional[str] = None,
            reasoning: Optional[str] = None,
        ) -> str:
            last_example: dict = start_last_example(
                helpfulness=helpfulness, reasoning=reasoning, pre_final="Excerpt 1"
            ) | dict(
                paragraphs=numbered_list(paragraphs),
                experiments=numbered_list(experiments),
                experiment_in_question=experiment_in_question,
            )
            shots = format_multi(
                CAN_WE_NAME_ARMS_TEMPLATE,
                CAN_WE_NAME_ARMS_EXAMPLES[:num_shot] + [last_example],
                SHARED_CAN_WE_NAME_ARMS_PROMPT,
            )
            instructions = """When evaluating a Randomized Controlled Trial, we should first identify its methodology; in particular, we should identify how different subgroups of participants were split into different trial arms, in order to receive different treatments or controls. We will look at a few paragraphs from different papers to identify whether any of those paragraphs identify the trial arms for specific experiments. Once we do this task, we will be able to draw a chart identifying, for each experiment, what the trial arms were. After that, we will look to identify how many participants were put into each arm, and how randomization was performed. Some of these excerpts may not be helpful, or it may be unclear. We'll try to find the paragraphs that most clearly identify what the different trial arms for the experiments in question were."""
            parts = [instructions] + list(shots)
            return "\n\n".join(parts)

        return make_can_we_name_arms_prompt

    return from_num_shot


CAN_WE_NAME_ARMS_CHOICES = [" Yes", " No", " Unsure"]
CAN_WE_NAME_ARMS_BEST_CHOICE = " Yes"
CAN_WE_NAME_ARMS_REASONING_STOP = ("\nWhich excerpt",)


def get_can_we_name_arms_reasoning(response: str) -> str:
    return "".join(
        ("Excerpt 1 ", get_part(response, "think it over:\n", "\nWhich excerpts"))
    )


def get_can_we_name_arms_helpfulness(response: str) -> str:
    return get_part(response, "conducted in this study?", "Therefore, do")



from ice.logging import init_logging

init_logging()

try:
    from .version import __version__
except ImportError:
    # version.py is auto-generated with the git tag when building
    __version__ = ""



from fvalues import F

from ice.recipes.primer.debate.utils import *


def render_debate_prompt(agent_name: str, debate: Debate, turns_left: int) -> str:
    prompt = F(
        f"""
You are {agent_name}. There are {turns_left} turns left in the debate. You are trying to win the debate using reason and evidence. Don't repeat yourself. No more than 1-2 sentences per turn.

{render_debate(debate, agent_name)}
You: "
"""
    ).strip()
    return prompt



from collections.abc import Sequence
from dataclasses import dataclass

from ice.formatter.transform.value import numbered_list


@dataclass
class Demonstration:
    question: str
    texts: Sequence[str]
    answer: str

    def as_dict(self):
        return dict(
            question=self.question, texts=numbered_list(self.texts), answer=self.answer
        )


@dataclass
class DemonstrationWithReasoning(Demonstration):
    reasoning: str

    def as_dict(self):
        return dict(
            question=self.question,
            texts=numbered_list(self.texts),
            answer=self.answer,
            reasoning=self.reasoning,
        )



import json

from os import environ
from pathlib import Path
from typing import Any
from typing import Optional
from typing import TYPE_CHECKING

from pydantic import BaseSettings
from structlog import get_logger

from .logging import log_lock

if TYPE_CHECKING:
    AnyHttpUrl = str
else:
    from pydantic import AnyHttpUrl

log = get_logger()


class Settings(BaseSettings):
    OPENAI_ORG_ID: str = ""
    OUGHT_INFERENCE_URL: AnyHttpUrl = "https://prod.elicit.org"
    GOLD_STANDARDS_CSV_PATH: Path = (
        Path(__file__).parent.parent / "gold_standards/gold_standards.csv"
    )
    GS_QUOTE_FOUND_THRESHOLD: float = 0.75
    OUGHT_ICE_HOST: str = "0.0.0.0"
    OUGHT_ICE_PORT: int = 8935
    OUGHT_ICE_AUTO_SERVER: bool = True
    OUGHT_ICE_AUTO_BROWSER: bool = True
    PAPER_DIR: Path = Path(__file__).parent.parent / "papers"

    # note these attributes are read differently- see [__getattribute__]
    OPENAI_API_KEY: str = ""
    OUGHT_INFERENCE_API_KEY: str = ""
    ELICIT_AUTH_TOKEN: str = ""

    def __get_and_store(self, setting_name: str, prompt: Optional[str] = None) -> str:
        # We use [__getattribute__] to read these attributes, so that we can
        # prompt the user for them if they are not already set.
        if prompt is None:
            prompt = f"Enter {setting_name}: "
        # We use this lock to prevent logging from another thread
        # from burying the input prompt. See [Trace._server_and_browser].
        with log_lock:
            value = input(prompt)
        setattr(self, setting_name, value)
        with open(_env_path, "a") as f:
            # [json.dumps] to escape quotes
            f.write(f"{setting_name}={json.dumps(value)}\n")
        log.info(f"Saved {setting_name} to {_env_path}")
        return value

    def __getattribute__(self, name: str) -> Any:
        result = super().__getattribute__(name)
        prompts = dict(
            OPENAI_API_KEY="Enter OpenAI API key (you can get this from https://beta.openai.com/account/api-keys): ",
            OUGHT_INFERENCE_API_KEY=None,
            ELICIT_AUTH_TOKEN="Enter Elicit Auth Token (you can find it by checking idToken in cookies for Elicit. "
            "The cookie should NOT start with Bearer; it should just be a string of letters and numbers): ",
        )
        if name in prompts and not result:
            return self.__get_and_store(name, prompts[name])
        else:
            return result


OUGHT_ICE_DIR = Path(environ.get("OUGHT_ICE_DIR", Path.home() / ".ought-ice"))

_env_path = OUGHT_ICE_DIR / ".env"

# Note that fields are loaded from pydantic in a particular priority ordering. See
# https://docs.pydantic.dev/usage/settings/#field-value-priority
settings = Settings(
    _env_file=_env_path if _env_path.exists() else None, _env_file_encoding="utf-8"
)

CACHE_DIR = OUGHT_ICE_DIR / "cache"
CACHE_DIR.mkdir(parents=True, exist_ok=True)


def server_url() -> str:
    return f"http://{settings.OUGHT_ICE_HOST}:{settings.OUGHT_ICE_PORT}"



from fvalues import F

Question = str
Answer = str
Subs = list[tuple[Question, Answer]]


def render_background(subs: Subs) -> str:
    if not subs:
        return ""
    subs_text = F("\n\n").join(F(f"Q: {q}\nA: {a}") for (q, a) in subs)
    return F(f"Here is relevant background information:\n\n{subs_text}\n\n")


def make_qa_prompt(question: str, subs: Subs) -> str:
    background_text = render_background(subs)
    return F(
        f"""{background_text}Answer the following question, using the background information above where helpful:

Question: "{question}"
Answer: "
"""
    ).strip()



import logging
import os
import sys
import threading

from structlog import configure
from structlog import get_logger
from structlog import make_filtering_bound_logger
from structlog import PrintLoggerFactory
from structlog.dev import ConsoleRenderer
from structlog.dev import set_exc_info
from structlog.processors import add_log_level
from structlog.processors import format_exc_info
from structlog.processors import TimeStamper
from structlog.types import Processor

# Prevent logging the warning 'None of PyTorch...' at the end of transformers/__init__.py.
# This depends on this being the first place that transformers is imported.
# It also assumes that transformers will be imported eventually
# so eagerly importing it now doesn't have an extra cost.
previous_verbosity = os.environ.get("TRANSFORMERS_VERBOSITY", None)
os.environ["TRANSFORMERS_VERBOSITY"] = "error"
import transformers  # noqa

# Allow using the TRANSFORMERS_VERBOSITY env var normally to still work,
# and avoid suppressing other warnings.
if previous_verbosity:  # Setting to None raises an error
    os.environ["TRANSFORMERS_VERBOSITY"] = previous_verbosity

# Used to prevent interleaved logging from multiple threads. See [Settings.__get_and_store].
log_lock = threading.Lock()


def init_logging():
    processors: list[Processor] = [
        add_log_level,
        TimeStamper(fmt="%Y-%m-%d %H:%M.%S.%f", utc=False),
        set_exc_info,
        format_exc_info,
        ConsoleRenderer(colors=sys.stdout is not None and sys.stdout.isatty()),
    ]
    configure(
        processors,
        cache_logger_on_first_use=True,
        logger_factory=PrintLoggerFactory(),
        wrapper_class=make_filtering_bound_logger(logging.INFO),
    )

    log = get_logger()
    log.debug("Logging initialized")



from collections.abc import Iterator
from collections.abc import Sequence
from math import factorial
from typing import TypeVar

import numpy as np

from structlog.stdlib import get_logger
from tqdm import tqdm

from ice.recipe import recipe
from ice.recipes.best_completion import completion_perplexity
from ice.utils import map_async

EXAMPLES_PROMPTS = [
    "Complete this math problem: 2 + 2 =",
    "Complete this math problem: 3 * 3 =",
    "Complete this math problem: 4 - 2 =",
]

EXAMPLES_COMPLETIONS = [
    " 4",
    " 9",
    " 2",
]

log = get_logger()


def remaining_prompts(
    prompts_and_completions: Sequence[tuple[str, str]],
    used_prompts_and_completions: Sequence[tuple[str, str]],
) -> Sequence[tuple[str, str]]:
    return [p for p in prompts_and_completions if p not in used_prompts_and_completions]


async def tuple_completion_perplexity(
    prompt_and_completion: tuple[str, str],
) -> float:
    return await completion_perplexity(*prompt_and_completion)


async def eval_prompt(
    few_shot_prompt: str,
    test_prompts_and_completions: list[tuple[str, str]],
    split_string: str,
) -> float:
    prompts = [
        few_shot_prompt + split_string + p for p, _ in test_prompts_and_completions
    ]

    completions = [completion for _, completion in test_prompts_and_completions]

    perplexities = await map_async(
        input_list=list(zip(prompts, completions)),
        fn=tuple_completion_perplexity,
        max_concurrency=10,
    )

    return sum(perplexities) / len(perplexities)


# Brake on going completely out of control here
MAX_PERMUTATIONS = 10_000

T = TypeVar("T")


def _random_permutations(
    lst: list[T], n_shots: int, rng: np.random.Generator, max=MAX_PERMUTATIONS
) -> Iterator[Sequence[T]]:
    for _ in range(max):
        copy = lst.copy()
        rng.shuffle(copy)  # type: ignore[arg-type]
        yield copy[:n_shots]


async def score_few_shot(
    examples_prompts: list[str] = EXAMPLES_PROMPTS,
    examples_completions: list[str] = EXAMPLES_COMPLETIONS,
    n_shots: int = 1,
    split_string: str = "\n\n",
    prefix: str = "",
    max_permutations: int = 10,
    max_test_size: int = 10,
    seed: int = 0,
) -> list[tuple[str, float]]:
    """Given some prompts divided into

    Args:
        examples_prompts (list[str], optional): _description_. Defaults to EXAMPLES_PROMPTS.
        examples_completions (list[str], optional): _description_. Defaults to EXAMPLES_COMPLETIONS.
        n_shots (int, optional): _description_. Defaults to 1.
        split_string (str, optional): _description_. Defaults to "\n\n".
        prefix (str, optional): _description_. Defaults to "".
        max_permutations (int, optional): _description_. Defaults to 10.
        max_test_size (int, optional): _description_. Defaults to 10.

    Returns:
        list[tuple[str, float]]: _description_
    """

    rng = np.random.default_rng(seed)

    assert len(examples_prompts) == len(examples_completions)

    n_possible_permutations = factorial(len(examples_prompts)) // factorial(
        len(examples_prompts) - n_shots
    )

    n_perms = min(n_possible_permutations, max_permutations)

    test_size = min(len(examples_prompts), max_test_size)

    log.info(
        "This will perform a brute force search of all possible combinations of few-shot prompts.",
        number_of_requests=n_perms * test_size,
    )

    prompts_and_completions = list(zip(examples_prompts, examples_completions))

    all_prompts: list[tuple[str, list[tuple[str, str]]]] = []

    for prompt_perm in _random_permutations(prompts_and_completions, n_shots, rng):
        prompt = prefix + split_string.join([p + c for p, c in prompt_perm])
        remaining_prompts_and_completions = remaining_prompts(
            prompts_and_completions, prompt_perm
        )

        test_prompt_completion_idxs = rng.choice(
            range(len(remaining_prompts_and_completions)), size=test_size, replace=False
        )

        test_prompts_and_completions = [
            remaining_prompts_and_completions[idx]
            for idx in test_prompt_completion_idxs
        ]

        all_prompts.append((prompt, test_prompts_and_completions))

    all_prompts_idxs = rng.choice(range(len(all_prompts)), size=n_perms, replace=False)

    all_prompts = [all_prompts[idx] for idx in all_prompts_idxs]

    scored_prompts = [
        (
            prompt,
            await eval_prompt(prompt, remaining_prompts_and_completions, split_string),
        )
        for prompt, remaining_prompts_and_completions in tqdm(
            all_prompts, desc="Evaluating prompts"
        )
    ]

    return scored_prompts  # The prompt with the lowest perplexity is the best few-shot prompt.


recipe.main(score_few_shot)



import re

from collections.abc import Iterator
from functools import cache
from pathlib import Path
from typing import Annotated
from typing import Literal
from typing import Optional

import nltk
import requests

from nltk.tokenize import sent_tokenize
from pydantic import BaseModel
from pydantic import Field
from structlog.stdlib import get_logger

from ice.cache import diskcache
from ice.environment import env
from ice.settings import OUGHT_ICE_DIR

log = get_logger()

PDF_PARSER_URL = "https://test.elicit.org/elicit-previews/james/oug-3083-support-parsing-arbitrary-pdfs-using/parse_pdf"

SectionType = Literal["abstract", "main", "back"]


@cache
def download_punkt():
    try:
        sent_tokenize("")
    except LookupError:
        nltk.download("punkt", quiet=False)


def get_full_document_id(document_id: str) -> str:
    return {
        "abebe-2018-tiny.txt": "abebe-2018.pdf",
        "keenan-2018-tiny.txt": "keenan-2018.pdf",
    }.get(document_id, document_id)


def get_paper_paths(paper_dir: Optional[Path] = None) -> list[Path]:
    if paper_dir is None:
        script_path = Path(__file__).parent.parent
        paper_dir = script_path / "papers/"
    paper_paths = [f for f in paper_dir.iterdir() if f.suffix in (".pdf", ".txt")]
    paper_paths = sorted(paper_paths)
    return paper_paths


def starts_with_number(s: str) -> bool:
    return re.match(r"^\d+", s) is not None


def is_likely_section_title(text: str):
    return (starts_with_number(text) and len(text) < 200) or text == "Abstract"


def split_sentences(text: str) -> list[str]:
    download_punkt()
    return sent_tokenize(text)


def section_title_words(title: str) -> str:
    if not starts_with_number(title):
        return title
    return " ".join(title.split(" ")[1:])


def section_title_number(title: str) -> str:
    if not starts_with_number(title):
        return ""
    return title.split(" ")[0]


def parse_txt(file: Path) -> list[dict]:
    body = []
    with open(file) as text_file:
        text = text_file.read()
        paragraphs = text.split("\n\n")
        current_section = "Introduction"
        for paragraph in paragraphs:
            if is_likely_section_title(paragraph):
                current_section = paragraph
            else:
                sentences = split_sentences(paragraph)
                body.append(
                    {
                        "sentences": sentences,
                        "sections": [
                            {
                                "title": section_title_words(current_section),
                                "number": section_title_number(current_section),
                            }
                        ],
                        "sectionType": "abstract"
                        if current_section == "Abstract"
                        else "main",
                    }
                )
    return body


def save_pdf_text(paper_body: list[dict], file_name: str):
    """
    Save pdf text to help with debugging.
    """
    paper_txt_dir = OUGHT_ICE_DIR / "paper_txt"
    paper_txt_dir.mkdir(parents=True, exist_ok=True)
    open(paper_txt_dir / f"{file_name}.txt", "w").write(
        "\n\n".join(" ".join(paragraph["sentences"]) for paragraph in paper_body)
    )


@diskcache()
def parse_pdf(file: Path) -> list[dict]:
    with env().spinner(f"[bold green] Parsing {file.name}"):
        files = {"pdf": open(file, "rb")}
        r = requests.post(PDF_PARSER_URL, files=files)
        body = r.json()
        save_pdf_text(body, file.name)
    env().print(f"Parsed {file.name}.")
    return body


class Section(BaseModel):
    title: str
    number: Optional[str] = None


class Paragraph(BaseModel):
    sentences: list[str]
    sections: list[Section]
    section_type: Annotated[SectionType, Field(alias="sectionType")]

    def is_body_paragraph(self):
        return self.section_type in ("abstract", "main")

    def is_empty(self):
        return str(self).strip() == ""

    def __str__(self):
        return " ".join(self.sentences)

    def __hash__(self):
        return hash(str(self))


class Paper(BaseModel):
    paragraphs: list[Paragraph]
    document_id: str = "unknown"

    @classmethod
    def load(cls, file: Path) -> "Paper":
        document_id = file.name

        if file.suffix == ".pdf":
            paragraph_dicts = parse_pdf(file)
        elif file.suffix == ".txt":
            paragraph_dicts = parse_txt(file)
        else:
            raise ValueError(f"Unknown extension: {file.suffix}")

        if len(paragraph_dicts) < 3:
            log.warn(f"paper {document_id} only has {len(paragraph_dicts)} paragraphs")

        return Paper.parse_obj(
            dict(paragraphs=paragraph_dicts, document_id=document_id)
        )

    def sentences(self) -> Iterator[str]:
        for paragraph in self.paragraphs:
            for sentence in paragraph.sentences:
                yield sentence

    def nonempty_paragraphs(self) -> list[Paragraph]:
        return [p for p in self.paragraphs if not p.is_empty()]

    def __str__(self):
        return "\n\n".join(str(p) for p in self.paragraphs)

    def dict(self, *args, **kwargs):
        kwargs["exclude"] = {"paragraphs"}
        return super().dict(*args, **kwargs)



from collections.abc import Sequence
from typing import ClassVar
from typing import Generic
from typing import Literal
from typing import NewType
from typing import Optional
from typing import Protocol
from typing import runtime_checkable
from typing import TypeVar
from typing import Union

from pydantic import BaseModel
from pydantic import root_validator
from pydantic import validator
from pydantic.generics import GenericModel

from ice.metrics.gold_standards import ParsedGoldStandardBase


T = TypeVar("T")

Stage = NewType("Stage", str)


class Sample(BaseModel):
    size: Optional[Union[int, Literal["Unclear"]]] = None
    stage: Optional[Stage] = None


class Arm(BaseModel):
    name: str
    description: str
    sample: Optional[Sample] = None

    @root_validator(pre=True)
    def fix_keys(cls, values):
        # Kind of ugly but the yaml in the airtable is in a pretty weird schema
        if isinstance(values, dict) and "arm description" in values:
            values["description"] = values["arm description"]
        if isinstance(values, dict) and "initial sample" in values:
            values["sample"] = values["initial sample"]
        return values


class Experiment(BaseModel):
    name: str
    description: str
    arms: list[Arm]

    @root_validator(pre=True)
    def fix_keys(cls, values):
        # Kind of ugly but the yaml in the airtable is in a pretty weird schema
        if isinstance(values, dict) and "experiment description" in values:
            values["description"] = values["experiment description"]
        return values

    @validator("arms", pre=True)
    def fix_arms(cls, v):
        # Kind of ugly but the yaml in the airtable is in a pretty weird schema
        if isinstance(v, dict):
            return [Arm.parse_obj(vals | dict(name=k)) for k, vals in v.items()]
        return v


class ExperimentsArms(ParsedGoldStandardBase):
    question_short_name: ClassVar[str] = "experiments_arms"
    experiments: list[Experiment]

    @validator("experiments", pre=True)
    def validate_experiments(cls, v):
        # Kind of ugly but the yaml in the airtable is in a pretty weird schema
        if isinstance(v, dict):
            return [Experiment.parse_obj(vals | dict(name=k)) for k, vals in v.items()]
        return v


ReasoningStage = Literal["reasoning", "helpfulness", "answer"]


class PassageWithReasoning(GenericModel, Generic[T]):
    passage: Sequence[str]
    reasoning: str
    helpfulness: Optional[str] = None
    score: Optional[float] = None
    final_answer: Optional[T] = None


@runtime_checkable
class MultipartReasoningPrompt(Protocol):
    def __call__(
        self,
        paragraphs: Sequence[str],
        helpfulness: Optional[str] = None,
        reasoning: Optional[str] = None,
    ) -> str:
        pass



import inspect

from typing import Type

import nest_asyncio
import pytest

from ice.recipe import Recipe
from ice.recipes import get_recipe_classes
from main import main_cli


nest_asyncio.apply()


def do_not_test(recipe: Type[Recipe]) -> bool:
    if hasattr(recipe, "do_not_test"):
        return recipe.do_not_test
    return False


def takes_paper_arg(recipe_class):
    run_signature = inspect.signature(recipe_class.run)
    return "paper" in run_signature.parameters


# Get recipe classes that take a paper argument in their run method
paper_recipe_classes = [
    recipe_class
    for recipe_class in get_recipe_classes()
    if not do_not_test(recipe_class) and takes_paper_arg(recipe_class)
]

# Get recipe classes that do not take a paper argument in their run method
no_paper_recipe_classes = [
    recipe_class
    for recipe_class in get_recipe_classes()
    if not do_not_test(recipe_class) and not takes_paper_arg(recipe_class)
]


@pytest.mark.parametrize(
    "recipe_name",
    [recipe_class.__name__ for recipe_class in paper_recipe_classes],
)
@pytest.mark.anyio
async def test_paper_recipes(recipe_name: str):
    main_cli(
        mode="test",
        input_files=["./papers/keenan-2018-tiny.txt"],
        recipe_name=recipe_name,
    )


@pytest.mark.parametrize(
    "recipe_name",
    [recipe_class.__name__ for recipe_class in no_paper_recipe_classes],
)
@pytest.mark.anyio
async def test_no_paper_recipes(recipe_name: str):
    main_cli(
        mode="test",
        recipe_name=recipe_name,
    )



import logging
import os
import re
import signal

from collections.abc import Awaitable
from collections.abc import Callable
from pathlib import Path

from fastapi import FastAPI
from fastapi import Request
from fastapi import Response
from fastapi.staticfiles import StaticFiles
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import FileResponse
from starlette.responses import PlainTextResponse

from ice.routes import traces
from ice.trace import traces_dir

logger = logging.getLogger(__name__)

dist_dir = Path(__file__).parent / "ui"


app = FastAPI()


# Add cache-control: no-transform header to all responses
# because otherwise cloudflare brotli compression breaks the trace viewer
async def add_no_transform_header(
    request: Request, call_next: Callable[[Request], Awaitable[Response]]
):
    response = await call_next(request)
    response.headers.update({"cache-control": "no-transform"})
    return response


app.add_middleware(BaseHTTPMiddleware, dispatch=add_no_transform_header)

app.include_router(traces.router)
app.mount(
    "/api/traces/", StaticFiles(directory=traces_dir), name="static"
)  # see comment on get_trace

try:
    app.mount("/assets/", StaticFiles(directory=dist_dir / "assets"), name="static")
except RuntimeError:
    logger.warning(
        "ui folder not found, skipping static file mount. Run `npm run build` to build the ui."
    )


PING_RESPONSE = "ought-ice says pong"


@app.get("/ping")
async def ping():
    return PlainTextResponse(PING_RESPONSE)


@app.post("/stop")
async def stop():
    # Note that this doesn't work properly when the server runs with --reload
    os.kill(os.getpid(), signal.SIGKILL)


@app.get("/{path:path}")
async def catch_all(path: str):
    # Never serve index.html for API requests or assets
    if re.match(r"^(?:api|assets)(?:/.*)?$", path):
        return PlainTextResponse("404 File Not Found", status_code=404)

    index_path = dist_dir / "index.html"
    if not index_path.exists():
        return PlainTextResponse(
            "ui/index.html not found. Run `npm run build` in the ui directory to create it, "
            "or run `npm run dev` to run the dev server and access localhost:5173 instead.",
            status_code=500,
        )

    return FileResponse(index_path)



import math  # noqa: F401

from abc import ABC
from abc import abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Optional

from fvalues import F

from ice.recipe import recipe
from ice.recipes.primer.search_string import search_string


Log = list[str]


def render_enumerate(items: Sequence[object]) -> str:
    """Render numbered list, one per line"""
    return F("\n\n").join(F(f"{i+1}. {item}") for i, item in enumerate(items))


def render_context(question: str, log: Log) -> str:
    question_context = F(f'The question you want to answer: "{question}"')
    if not log:
        return question_context
    return F(
        f"""{question_context}

What you've done so far:

{render_enumerate(log)}"""
    )


def render_action_context(question: str, log: Log, max_actions: int) -> str:
    action_count_text = (
        "You have one action left. (The one you're taking right now.)"
        if max_actions == 1
        else F(
            f"You have {max_actions} actions left. (The one you're taking right now, and {max_actions - 1} follow-up actions.)"
        )
    )
    return F(
        f"""{render_context(question, log)}

{action_count_text}"""
    )


def make_knowledge_prompt(question: str, log: Log) -> str:
    return F(
        f"""{render_context(question, log)}

Q: Do you have enough information to correctly answer the question? Say "A: Yes" or "A: No"
A:"""
    )


def make_answer_prompt(question: str, log: Log) -> str:
    return F(
        f"""{render_context(question, log)}

Q: {question}
A:"""
    )


async def is_info_sufficient(question: str, log: Log) -> bool:
    knowledge_prompt = make_knowledge_prompt(question, log)
    has_knowledge_probs, _ = await recipe.agent().classify(
        prompt=knowledge_prompt, choices=(" Yes", " No")
    )
    return has_knowledge_probs.get(" Yes", 0.0) > 0.7


async def answer_directly(question: str, log: Log) -> str:
    answer_prompt = make_answer_prompt(question, log)
    answer = await recipe.agent("instruct-reasoning").complete(prompt=answer_prompt)
    return answer


class Action(ABC):
    @classmethod
    @abstractmethod
    async def propose(cls, question: str, log: Log, max_actions: int) -> "Action":
        ...

    @abstractmethod
    def run(self):
        ...

    @abstractmethod
    def make_log_entry(self, result: str) -> str:
        ...


@dataclass
class CalculationAction(Action):
    calculation: str

    @classmethod
    def make_proposal_prompt(cls, question: str, log: Log, max_actions: int) -> str:
        return F(
            f"""{render_action_context(question, log, max_actions)}

You have chosen to take the action "Do a calculation".

You have access to a Python interpreter. What single-line calculation would most help you answer the question "{question}"?

>>> import math
>>>"""
        )

    @classmethod
    async def propose(
        cls, question: str, log: Log, max_actions: int
    ) -> "CalculationAction":
        calculation_prompt = cls.make_proposal_prompt(question, log, max_actions)
        calculation = await recipe.agent("instruct-reasoning").complete(
            prompt=calculation_prompt, stop="\n"
        )
        return cls(calculation)

    async def run(self) -> str:
        try:
            return str(eval(self.calculation))
        except Exception as e:
            return F(f"Error: {e}")

    def make_log_entry(self, result: str) -> str:
        return F(f"You calculated '{self.calculation}' and got the result '{result}'.")

    def __str__(self):
        return F(f"Do calculation: {self.calculation}")


@dataclass
class WebSearchAction(Action):
    search_term: str

    @classmethod
    def make_proposal_prompt(cls, question: str, log: Log, max_actions: int) -> str:
        return F(
            f"""{render_action_context(question, log, max_actions)}

You have chosen to take the action "Run a web search".

What is a first web search query you could run to help you answer the question "{question}"?

Query:"""
        )

    @classmethod
    async def propose(
        cls, question: str, log: Log, max_actions: int
    ) -> "WebSearchAction":
        search_term_prompt = cls.make_proposal_prompt(question, log, max_actions)
        search_term = await recipe.agent("instruct-reasoning").complete(
            prompt=search_term_prompt, stop='"'
        )
        return cls(search_term)

    async def run(self) -> str:
        results_str = await search_string(self.search_term)
        return results_str

    def make_log_entry(self, result: str) -> str:
        return F(
            f"You searched the web for '{self.search_term}' and got the result '{result}'."
        )

    def __str__(self):
        return F(f"Run web search: {self.search_term}")


async def get_action_candidates(
    question: str, log: Log, max_actions: int
) -> list[Action]:
    calculation_action = await CalculationAction.propose(question, log, max_actions)
    websearch_action = await WebSearchAction.propose(question, log, max_actions)
    return [calculation_action, websearch_action]


def render_numbers(n: int) -> str:
    numbers = ", ".join(str(i) for i in range(1, n))
    return "{} or {}".format(numbers, n) if n > 1 else str(n)


def make_action_choice_prompt(
    question: str, log: Log, actions: list[Action], max_actions: int
) -> str:
    follow_up_text = (
        ""
        if max_actions == 1
        else F(f", and {max_actions - 1} similar follow-up actions")
    )

    return F(
        f"""{render_context(question, log)}

You can take one of the following actions now{follow_up_text} before you need to answer:

{render_enumerate(actions)}

Question: What next action should you take to make progress on answering the question "{question}"? {render_numbers(len(actions))}?
Answer:"""
    )


async def choose_action(
    question: str, log: Log, actions: list[Action], max_actions
) -> Action:
    action_choice_prompt = make_action_choice_prompt(
        question, log, actions, max_actions
    )
    action_choice_probs = await get_action_choice_probs(action_choice_prompt, actions)
    best_action_index = get_best_action_index(action_choice_probs)
    return actions[best_action_index]


async def get_action_choice_probs(action_choice_prompt, actions):
    action_choice_probs, _ = await recipe.agent("instruct-reasoning-crowd").classify(
        prompt=action_choice_prompt,
        choices=tuple(F(f" {i}") for i in range(1, len(actions) + 1)),
    )
    return action_choice_probs


def get_best_action_index(action_choice_probs):
    best_action_index = None
    best_action_prob = 0
    for action_index, action_prob in action_choice_probs.items():
        if action_prob > best_action_prob:
            best_action_index = int(action_index.strip())
            best_action_prob = action_prob
    assert best_action_index is not None
    return best_action_index - 1


async def gather_info(
    *,
    question: str,
    log: Optional[Log] = None,
    max_actions: int = 3,
) -> Log:
    if log is None:
        log = []
    actions = await get_action_candidates(question, log, max_actions)
    chosen_action = await choose_action(question, log, actions, max_actions)
    result = await chosen_action.run()
    return log + [chosen_action.make_log_entry(result)]


async def sequential_action(
    *,
    question: str = "How far would all the film frames that make up the 400-plus episodes of The Simpsons stretch?",
    max_actions: int = 3,
):
    log: list[str] = []

    for actions_left in range(max_actions, 0, -1):

        sufficient_info = await is_info_sufficient(question, log)
        if sufficient_info:
            break

        log = await gather_info(
            question=question,
            log=log,
            max_actions=actions_left,
        )

    answer = await answer_directly(question, log)

    return answer


recipe.main(sequential_action)



from fvalues import F

from ice.recipe import recipe


def eval_python(expression: str) -> str:
    try:
        result = eval(expression)
    except Exception as e:
        result = F(f"Error: {e}")
    return str(result)


async def answer_by_computation(question: str):
    return eval_python(question)


recipe.main(answer_by_computation)



"""
Make a dataframe that contains the paragraphs that contain the gold standard quotes.
"""
import asyncio

from pathlib import Path
from typing import Optional

import pandas as pd

from structlog import get_logger

from ice.cache import diskcache
from ice.evaluation.utils import rouge_compare
from ice.metrics.gold_standards import get_gold_standards
from ice.metrics.gold_standards import list_experiments
from ice.paper import get_paper_paths
from ice.paper import Paper
from ice.paper import Paragraph

log = get_logger()


def get_question_gold_standards(question_short_name: str):
    paper_paths = get_paper_paths()
    question_gold_standards = []
    for paper_path in paper_paths:
        paper = Paper.load(paper_path)
        experiments = list_experiments(
            document_id=paper.document_id, question_short_name=question_short_name
        )
        for experiment in experiments:
            gold_standards = get_gold_standards(
                document_id=paper.document_id,
                question_short_name=question_short_name,
                experiment=experiment,
            )
            if gold_standards:
                assert len(gold_standards) == 1
                gold_standard = gold_standards[0]
                question_gold_standards.append(gold_standard)
    return question_gold_standards


def nonempty_paragraphs(paper):
    for paragraph in paper.paragraphs:
        if paragraph.sentences:
            yield paragraph


def get_recall(quote, paragraph):
    rouge = asyncio.run(rouge_compare([str(paragraph.sentences)], [quote]))
    recall = rouge.rouge_l.r
    return recall


@diskcache()
def get_containing_paragraph(
    document_id: str, quote: str, verbose: bool = False
) -> Optional[Paragraph]:
    """
    Given a document ID and a quote, return the paragraph that the quote is in.
    """
    script_path = Path(__file__).parent
    paper = Paper.load(script_path / ".." / ".." / "papers" / document_id)
    best_recall_paragraph = None
    best_recall = 0.0
    for paragraph in nonempty_paragraphs(paper):
        recall = get_recall(quote, paragraph)
        if recall > best_recall:
            best_recall_paragraph = paragraph
            best_recall = recall
    if best_recall < 0.75 and verbose:
        # Explanations:
        # - Quote is split across two paragraphs
        # - Document paragraphs don't include quote
        log.warning(
            f"""Couldn't find gold standard paragraph for quote

> {quote}

in {document_id}. Best recall was {best_recall:.2f}. Best paragraph was:

> {best_recall_paragraph}"""
        )
    return best_recall_paragraph


def get_gold_paragraph_df(question_short_name: str):

    gold_standards = get_question_gold_standards(question_short_name)

    entries = []
    paragraph_id = 0
    id_to_paragraph = {}

    for gold_standard in gold_standards:
        base_entry = {
            "document_id": gold_standard.document_id,
            "experiment": gold_standard.experiment,
            "answer": gold_standard.answer,
            "classification": gold_standard.classifications[1],
        }
        for quote in gold_standard.quotes:
            entry = base_entry.copy()
            entry["quote"] = quote
            containing_paragraph = get_containing_paragraph(
                entry["document_id"], entry["quote"]
            )
            entry["paragraph"] = str(containing_paragraph)
            entry["paragraph_id"] = paragraph_id
            id_to_paragraph[paragraph_id] = containing_paragraph
            entries.append(entry)
            paragraph_id += 1

    # Convert entries to dataframe
    df = pd.DataFrame(entries)

    # Reorder columns, classification last
    df = df[
        [
            "document_id",
            "experiment",
            "quote",
            "paragraph",
            "paragraph_id",
            "answer",
            "classification",
        ]
    ]

    df = df.rename(
        columns={
            "answer": "paper_gold_answer",
            "classification": "paper_gold_classification",
        }
    )
    return df, id_to_paragraph



from ice.recipe import recipe
from ice.recipes.abstract_qa import Abstract
from ice.recipes.abstract_qa import DEFAULT_ABSTRACTS


def make_prompt(question: str, abstracts: list[Abstract], answers: list[str]) -> str:

    abstract_answers_str = "\n\n".join(
        [
            f"Title B{i}: {abstract.title}\nAbstract B{i}: {abstract.text}\nAnswer B{i}: {answer}"
            for i, (abstract, answer) in enumerate(zip(abstracts, answers), start=1)
        ]
    )

    return f"""For each set of abstracts, combine the abstract-wise answers to answer the question. Provide references to the original answers to support each claim in your combined answer.

###

Question A: What is IL-13?

Abstract A1: Interleukin 13 (IL-13) is a recently described protein secreted by activated T cells which is a potent in vitro modulator of human monocyte and B-cell functions. The data, reviewed here by Gerard Zurawski and Jan de Vries, shows that IL-13 shares biological activities with IL-4, their genes are closely linked in both the human and mouse genomes, and there is sequence homology between IL-13 and IL-4 proteins. Although the cloned IL-4 receptor protein (IL-4R) does not bind IL-13, it appears that the functional IL-4R and IL-13R share a common subunit that is important for signal transduction.

Answer A1: Interleukin 13 (IL-13) is a recently described protein secreted by activated T cells which is a potent in vitro modulator of human monocyte and B-cell functions. It is similar to IL-4.

Abstract A2: Interleukin-13 (IL-13), like IL-4, is a cytokine produced by TH2 type helper T cells in response to signaling through the T cell antigen receptor and by mast cells and basophils upon cross-linkage of the high-affinity receptor for immunoglobulin E (IgE). It is also produced by activated eosinophils. IL-13 induces many of the same responses as IL-4 and shares a receptor subunit with IL-4. IL-13 has been implicated in airway hypersensitivity and mucus hypersecretion, inflammatory bowel disease, and parasitic nematode expulsion.

Answer A2: Interleukin-13 (IL-13) is a cytokine produced by TH2 type helper T cells, mast cells, basophils, and eosinophils. It has been implicated in airway hypersensitivity and mucus hypersecretion, inflammatory bowel disease, and parasitic nematode expulsion. It is similar to IL-4.

Abstract A3: The recently cloned human interleukin 13 (IL-13) is a novel cytokine expressed in activated T cells that has been shown to inhibit inflammatory cytokine production by lipopolysaccharide-activated monocytes. The protein encoded by the IL-13 cDNA is the human homologue of a mouse Th2-product called P600. Here, we show that IL-13 acts at different stages of the B cell maturation pathway: (a) it enhances the expression of CD23/Fc epsilon RII and class II MHC antigens on resting B cells; (b) it stimulates B cell proliferation in combination with anti-Ig and anti-CD40 antibodies; and (c) it induces IgE synthesis. Thus, the spectrum of the biological activities of IL-13 on B cells largely overlaps that previously ascribed to IL-4. The present observations suggest that IL-13 may be an important factor, in addition to IL-4, in the development of allergic diseases.

Answer A3: Interleukin 13 (IL-13) is a novel cytokine expressed in activated T cells that has been shown to inhibit inflammatory cytokine production. It does similar things to IL-4 and may be an important factor in the development of allergic diseases.

Abstract A4: The discovery of new cytokines normally relies on a prior knowledge of at least one of their biological effects, which is used as a criterion either for the purification of the protein or for the isolation of the complementary DNA by expression cloning. However, the redundancy of cytokine activities complicates the discovery of novel cytokines in this way, and the pleiotropic nature of many cytokines means that the principal activities of a new cytokine may bear little relation to that used for its isolation. We have adopted an alternative approach which relies on differential screening of an organized subtracted cDNA library from activated peripheral blood mononuclear cells, using the inducibility of lymphokine messenger RNAs by anti-CD28 as a primary screening criterion. The ligation of the CD28 antigen on the T lymphocyte by a surface antigen, B7/BB-1, expressed on activated B lymphocytes and monocytes is a key step in the activation of T lymphocytes and the accumulation of lymphokine mRNAs. Here we report the discovery by molecular cloning of a new interleukin (interleukin-13 or IL-13) expressed in activated human T lymphocytes. Recombinant IL-13 protein inhibits inflammatory cytokine production induced by lipopolysaccharide in human peripheral blood monocytes. Moreover, it synergizes with IL-2 in regulating interferon-gamma synthesis in large granular lymphocytes. Recent mapping of the IL-13 gene shows that it is closely linked to the IL-4 gene on chromosome 5q 23-31 (ref. 4). Interleukin-13 may be critical in regulating inflammatory and immune responses.

Answer A4: Recombinant Interleukin-13 (IL-13) protein inhibits inflammatory cytokine production induced by lipopolysaccharide in human peripheral blood monocytes. Moreover, it synergizes with IL-2 in regulating interferon-gamma synthesis in large granular lymphocytes.

Combined Answer A with references to A1-A4: IL-13 is a cytokine protein secreted by T cells, mast cells, basophils, and eosinophils in humans (A1; A2; A3). It modulates monocyte and B-cell functions and inhibits imflmmatory cytokine production (A1; A3; A4) and is implicated in airway hypersensitivity and mucus hypersecretion, inflammatory bowel disease, and parasitic nematode expulsion (A2). It is similar to IL-4 (A1; A2; A3).

###

Question B: {question}

{abstract_answers_str}

Combined Answer B with references to B1-B4:"""


async def combine_abstract_answers(
    question: str, abstracts: list[Abstract], answers: list[str]
) -> str:
    prompt = make_prompt(question=question, abstracts=abstracts, answers=answers)
    answer = await recipe.agent().complete(prompt=prompt, stop="###")
    return answer


async def combine_abstract_answers_cli() -> str:
    abstracts = DEFAULT_ABSTRACTS
    question = "what is the relationship between income and smoking?"
    answers = [
        "Studies worldwide show a consistent inverse dose-response relationship between cigarette smoking and income level, present among most geographical areas and country characteristics.",
        "Income and education are both related to smoking in the EU",
        "There is an inverse relationship between income and smoking prevalence in studies from Latin America",
        "Education and income are both negatively related to smoking in health surveys in Hungary",
    ]
    return await combine_abstract_answers(
        question=question, abstracts=abstracts, answers=answers
    )


recipe.main(combine_abstract_answers_cli)



from collections.abc import Sequence
from typing import Optional
from typing import Union

from ice.formatter.multi import format_multi
from ice.formatter.multi import stop
from ice.formatter.transform.value import numbered_list
from ice.formatter.transform.value import ValueTransform
from ice.recipes.experiments_and_arms.prompts.utils import get_part
from ice.recipes.experiments_and_arms.prompts.utils import start_last_example
from ice.recipes.experiments_and_arms.types import MultipartReasoningPrompt

CAN_WE_COUNT_EXPERIMENTS_EXAMPLE_TEMPLATE = """Do the following excerpts explain the methodology of the paper, including how many experiments were conducted, and what the trial arms in each experiment were? If the passages merely discuss multiple interventions but do not specify that they belong to distinct experiments, answer "Unclear".

Excerpts:

{paragraphs}

Let's think it over:

{reasoning}

Which excerpts, if any, were helpful in understanding how many experiments were conducted in this study? {helpfulness}

Therefore, do the above excerpts explain the methodology of the paper, including how many experiments were conducted, and what the trial arms in each experiment were? If we cannot tell from the excerpts, answer "Unclear". Answer Yes, No, or Unclear.

Answer: {answer}""".strip()


CAN_WE_COUNT_EXPERIMENTS_EXAMPLES: list[
    dict[str, Union[ValueTransform[Sequence[str]], str]]
] = [
    dict(
        paragraphs=numbered_list(
            [
                """We find that the effects of the workshop are significantly larger for the most disadvantaged. In particular, we show in Table 5 that the least educated, the least experienced, and those with the lowest expected earnings benefit the most from the interventions. For other dimensions, we are unable to find significant differences in response to treatment. 51  he size of the effects for the worst-off workers is substantial. For example, young people without tertiary education increase the earnings by almost 60 percent, while the low predicted earnings group experiences a 50 percent increase. This causes a large reduction in earning inequality: the earning gap between the low and the high earnings group drops from 142 percent to 54 percent and, strikingly, the gap between experienced and inexperienced workers is fully erased. Overall, these results illustrate the large equity gains that can be generated by helping young workers to access the labour market through improved signalling.""",
                """A.17""",
                """Note. In this table we report the intent-to-treat estimates of the direct and indirect effects of the transport intervention and the job application workshop on financial outcomes. These are obtained by OLS estimation of equation ( 1), weighting each observation by the inverse of the probability of being sampled. Below each coefficient estimate, we report the s.e. in parentheses and the q-value in brackets. We correct standard errors to allow for arbitrary correlation at the level of geographical clusters. q-values are obtained using the sharpened procedure of Benjamini et al. (2006). Changing number of observations due to missing values in the dependent variable. In the last three columns we report the mean outcome for the control group, the p-value from a F-test of the null hypothesis that transport subsidies and the job application workshop have the same effect, and the number of observations. ***p< 0.01, **p<0.05, *p<0.1. Note. In this table we report the intent-to-treat estimates of the direct and indirect effects of the transport intervention and the job application workshop on expectations, aspirations and reservation wages. These are obtained by OLS estimation of equation ( 1), weighting each observation by the inverse of the probability of being sampled. Below each coefficient estimate, we report the s.e. in parentheses and the q-value in brackets. We correct standard errors to allow for arbitrary correlation at the level of geographical clusters. q-values are obtained using the sharpened procedure of Benjamini et al. (2006). Changing number of observations due to missing values in the dependent variable. In the last three columns we report the mean outcome for the control group, the p-value from a F-test of the null hypothesis that transport subsidies and the job application workshop have the same effect, and the number of observations. ***p< 0.01, **p<0.05, *p<0.1. Note. In this table we report the intent-to-treat estimates of the direct and indirect effects of the transport intervention and the job application workshop on outcomes related to mobility. These are obtained by OLS estimation of equation ( 1), weighting each observation by the inverse of the probability of being sampled. Below each coefficient estimate, we report the s.e. in parentheses and the q-value in brackets. We correct standard errors to allow for arbitrary correlation at the level of geographical clusters. q-values are obtained using the sharpened procedure of Benjamini et al. (2006). Changing number of observations due to missing values in the dependent variable. In the last three columns we report the mean outcome for the control group, the p-value from a F-test of the null hypothesis that transport subsidies and the job application workshop have the same effect, and the number of observations. ***p< 0.01, **p<0.05, *p<0.1.""",
                """A.19 1), weighting each observation by the inverse of the probability of being sampled. Below each coefficient estimate, we report the s.e. in parentheses and the q-value in brackets. We correct standard errors to allow for arbitrary correlation at the level of geographical clusters. q-values are obtained using the sharpened procedure of Benjamini et al. (2006). Changing number of observations due to missing values in the dependent variable. In the last three columns we report the mean outcome for the control group, the p-value from a F-test of the null hypothesis that transport subsidies and the job application workshop have the same effect, and the number of observations. ***p< 0.01, **p<0.05, *p<0.1.""",
            ]
        ),
        reasoning="""Excerpt 1 discusses findings related to a workshop; this appears to be from a section of the paper reporting results, not a methodology section, and it's not clear whether this paper conducted multiple experiments or just one.

Excerpt 2 is just a reference and does not have any useful content.

Excerpt 3 is a note explaining some analytic methodology related to reporting outcomes from the job application workshop and the transport subsidy. However, it does not clarify whether these belong to separate experiments.

Excerpt 4 also explains some analytic methodology; because it seems to compare the effect sizes of the job application workshop and the transport subsidy together, it suggests that these might be two trial arms belonging to a single experiment, but it is ultimately unclear here.""",
        helpfulness="""Excerpts 3 and 4 were somewhat helpful. Excerpts 1 and 2 were not helpful.""",
        answer="Unclear",
    )
]


def make_can_we_count_experiments_prompt(num_shots: int) -> MultipartReasoningPrompt:
    def can_we_count_experiments_prompt(
        paragraphs: Sequence[str],
        helpfulness: Optional[str] = None,
        reasoning: Optional[str] = None,
    ) -> str:
        last_example = start_last_example(
            helpfulness=helpfulness,
            reasoning=reasoning,
            pre_final="Excerpt 1",
        )
        last_example = last_example | dict(
            paragraphs=numbered_list(paragraphs),
            reasoning=reasoning if reasoning else stop("Excerpt 1"),
        )

        prompt = "\n\n".join(
            format_multi(
                CAN_WE_COUNT_EXPERIMENTS_EXAMPLE_TEMPLATE,
                CAN_WE_COUNT_EXPERIMENTS_EXAMPLES[:num_shots] + [last_example],
            )
        )
        return prompt

    return can_we_count_experiments_prompt


def get_helpfulness(response: str):
    return get_part(
        response,
        "understanding how many experiments were conducted in this study? ",
        "\n\n",
    )


def get_reasoning(response: str):
    return "".join(
        ("Excerpt 1 ", get_part(response, "Let's think it over:", "\nWhich excerpt"))
    )


CAN_WE_COUNT_EXPERIMENTS_CHOICES = [" Yes", " No", " Unclear"]
CAN_WE_COUNT_EXPERIMENTS_BEST_CHOICE = " Yes"

CAN_WE_COUNT_EXPERIMENTS_REASONING_STOP = ("\n\nWhich excerpt",)



from fvalues import F

from ice.recipe import recipe


DEFAULT_CONTEXT = "We're running a hackathon on 9/9/2022 to decompose complex reasoning tasks into subtasks that are easier to automate & evaluate with language models. Our team is currently breaking down reasoning about the quality of evidence in randomized controlled trials into smaller tasks e.g. placebo, intervention adherence rate, blinding procedure, etc."

DEFAULT_QUESTION = "What is happening on 9/9/2022?"


def make_qa_prompt(context: str, question: str) -> str:
    return F(
        f"""
Background text: "{context}"

Answer the following question about the background text above:

Question: "{question}"
Answer: "
"""
    ).strip()


async def answer(
    context: str = DEFAULT_CONTEXT, question: str = DEFAULT_QUESTION
) -> str:
    prompt = make_qa_prompt(context, question)
    answer = await recipe.agent().complete(prompt=prompt, stop='"')
    return answer


recipe.main(answer)



from ice.recipe import recipe
from ice.recipes.abstract_qa import Abstract
from ice.recipes.elicit.search import elicit_search
from ice.recipes.synthesize import synthesize
from ice.recipes.synthesize_compositional import synthesize_compositional


async def search_and_synthesize(
    question: str = "What is the effect of creatine on cognition?",
    compositional: bool = False,
):
    elicit_response = await elicit_search(question=question, num_papers=4)
    elicit_papers = list(elicit_response["papers"].values())

    abstracts = [
        Abstract(
            title=paper["title"],
            authors=paper["authors"],
            year=paper["year"],
            text=paper["unsegmentedAbstract"],
        )
        for paper in elicit_papers
    ]

    if compositional:
        synthesis = await synthesize_compositional(
            question=question, abstracts=abstracts
        )
    else:
        synthesis = await synthesize(question=question, abstracts=abstracts)

    return synthesis


recipe.main(search_and_synthesize)



from ice.recipe import recipe
from ice.recipes.primer.answer_by_dispatch.prompt import *


async def select_action(question: str) -> Action:
    prompt = make_action_selection_prompt(question)
    choices = tuple(str(i) for i in range(1, 6))
    choice_probs, _ = await recipe.agent().classify(prompt=prompt, choices=choices)
    best_choice = max(choice_probs.items(), key=lambda x: x[1])[0]
    return action_types[int(best_choice) - 1]


async def answer_by_dispatch(question: str = "How many people live in Germany?") -> str:
    action = await select_action(question)
    result = await action.recipe(question=question)
    return result


recipe.main(answer_by_dispatch)



from fvalues import F

from ice.json_value import to_json_value
from ice.summarize import summarize


def test_f():
    name = "world"
    f = F(f"hello {name}")
    assert to_json_value(f) == {
        "__fstring__": [
            "hello ",
            {"source": "name", "value": "world", "formatted": "world"},
        ]
    }
    assert summarize(to_json_value(f)) == {"name": "world"}
    f += "!"
    assert summarize(to_json_value(f)) == {"f": {"name": "world"}}


def test_basic():
    assert summarize(list(range(10))) == [0, 1, 2]
    d: dict = dict(enumerate(range(100)))
    assert summarize(d) == {
        0: 0,
        1: 1,
        2: 2,
        3: 3,
        4: 4,
        5: 5,
        6: 6,
        7: 7,
        8: 8,
        9: 9,
    }
    assert (
        summarize("abc" * 100)
        == "abcabcabcabcabcabcabcabcabcabcabcabcabcabcabcabc...cabcabcabcabcabcabcabcabcabcabcabcabcabcabcabcabc"
    )


def test_depth():
    assert summarize({"a": {"b": {"c": 3}}}) == {
        "a": {
            "b": {
                "c": 3,
            }
        }
    }
    assert (
        summarize(
            {
                "a": {
                    "b": {
                        "c": {
                            "d": 3,
                        }
                    }
                }
            }
        )
        == {}
    )
    assert summarize({"a": {"b": {"c": {"d": 3}, "e": 4}}}) == {
        "a": {
            "b": {
                "e": 4,
            },
        }
    }


def test_empty():
    assert summarize({"a": "b", "c": None, "d": {}, "e": [], "f": ""}) == {
        "a": "b",
    }


def test_complex():
    assert summarize(
        {
            "a": {
                "b": [1, 2, 3, 4, 5],
                "c": {
                    "d": "abc" * 100,
                    "e": 1.23456789,
                },
            },
            "f": None,
            "g": {
                "h": {
                    "i": {
                        "j": {
                            "k": "hello",
                        }
                    }
                }
            },
        }
    ) == {
        "a": {
            "b": [1, 2, 3],
            "c": {
                "d": "abcabcabcabcabcabcabcabcabcabcabcabcabcabcabcabc...cabcabcabcabcabcabcabcabcabcabcabcabcabcabcabcabc",
                "e": 1.2346,
            },
        }
    }



import pytest

from ice.formatter.transform.dependent import CountWord
from ice.formatter.transform.dependent import DependentTransform
from ice.formatter.transform.dependent import plural_transform


@pytest.mark.parametrize(("case", "expected"), (((1, 2, 3), "are"), (1, "is")))
def test_plural_transform(case, expected):
    xform = plural_transform("key", singular_case="is", plural_case="are")
    assert isinstance(xform, DependentTransform)
    assert xform.key() == "key"
    assert xform.transform(case) == expected


def test_count_word():
    xform = CountWord("key", special={4: "yay four"})
    for case, expected in (
        (1, "one"),
        (["hi", "hello"], "two"),
        (3, "three"),
        ((0, 0, 0, 0), "yay four"),
    ):
        assert xform.transform(case) == expected  # type: ignore[arg-type]



import pytest

from structlog.testing import capture_logs

from ice.formatter.multi import _parse
from ice.formatter.multi import _STOP_EARLY_WARNING
from ice.formatter.multi import _unparse
from ice.formatter.multi import format_multi
from ice.formatter.multi import stop
from ice.formatter.transform.dependent import CountWord
from ice.formatter.transform.positional import OrdinalWord
from ice.formatter.transform.value import non_literal
from ice.formatter.transform.value import numbered_list


def test_parse_and_unparse_are_inverses():
    complex_format_string = "{{This is some {{nested}} nonsense!}}, \
        and {{conversions!r}} and empty {{:2f}} {{}} {{!a:2d}} format \
        strings work as well as unmatched and escaped {{}} \
        }} braces {{ right up until the {{end:.2f}}"
    assert _unparse(_parse(complex_format_string)) == complex_format_string


def test_stop_partial():
    format_str = "We can start {here} and keep going {there}."
    cases = format_multi(
        format_str,
        [dict(here="in Egypt", there="to France"), dict(here=stop("partying."))],
    )
    assert cases == (
        "We can start in Egypt and keep going to France.",
        "We can start partying.",
    )


def test_warn_stop_early():
    format_str = "We can start {here} and keep going {there}."
    with capture_logs() as logs:
        cases = format_multi(
            format_str,
            [
                dict(here=stop("partying."), there="to France"),
                dict(here="in Egypt", there="to France"),
            ],
        )
    events = {logged["event"] for logged in logs}
    assert _STOP_EARLY_WARNING in events
    assert cases == (
        "We can start partying.",
        "We can start in Egypt and keep going to France.",
    )


def test_shared():
    format_str = "{a} and {b}"
    differing = [dict(a="pizza"), dict(a="hotdogs")]
    shared = dict(b="fries")
    cases = format_multi(format_str, differing, shared)
    assert cases == ("pizza and fries", "hotdogs and fries")

    overriding_case = [dict(a="pizza", b="ice cream")]
    overriden = format_multi(format_str, overriding_case, shared)[0]
    assert overriden == "pizza and ice cream"


def test_apply_transforms():
    format_str = "{a} {b}"
    special = [1, 2, 3]
    with pytest.raises(KeyError):
        format_multi(format_str, [dict(a=non_literal(special))])

    special_count = CountWord("a")

    filled = format_multi(format_str, [dict(a=non_literal(special), b=special_count)])[
        0
    ]
    assert filled == f"{special} three"

    filled_numbered_list = format_multi(
        format_str,
        [
            dict(
                a=numbered_list([str(val) for val in special], separator="..."),
                b=special_count,
            )
        ],
    )[0]
    assert filled_numbered_list == "1. 1...2. 2...3. 3 three"

    filled_positions = format_multi(
        format_str, [dict(b="this"), dict(b="that")], dict(a=OrdinalWord())
    )
    assert filled_positions == ("first this", "second that")



from typing import Type

from ice.recipe import Recipe

from .adherence_keyword_baseline import AdherenceKeywordBaseline
from .adherence_simple import AdherenceSimpleInstruct
from .adherence_tfew_paragraph import AdherenceParagraphTfew
from .all_quotes import AllQuotesRecipe
from .blinding_dynamic import BlindingDynamic
from .comparisons_qa import ComparisonsQA
from .evaluate_result import EvaluateResult
from .evaluate_results import EvaluateResults
from .experiment_arms import ExperimentArms
from .experiments_and_arms.recipes.experiments_and_arms import ExperimentsAndArms
from .experiments_and_arms.recipes.name_experiments import NameExperiments
from .funnel_simple import FunnelSimple
from .placebo_description import PlaceboDescriptionInstruct
from .placebo_dialogs import PlaceboDialogs
from .placebo_keyword_baseline import PlaceboKeywordBaseline
from .placebo_simple import PlaceboSimpleInstruct
from .placebo_tree import PlaceboTree
from .program_search.nodes.decontext.decontextualize import PaperDecontext
from .rank_paragraphs import RankParagraphs
from .subrecipe_example import ExampleMetaRecipe


def get_recipe_classes() -> list[Type[Recipe]]:
    return [
        AdherenceKeywordBaseline,
        AdherenceParagraphTfew,
        AdherenceSimpleInstruct,
        AllQuotesRecipe,
        BlindingDynamic,
        ComparisonsQA,
        NameExperiments,
        ExperimentsAndArms,
        EvaluateResult,
        EvaluateResults,
        ExampleMetaRecipe,
        PaperDecontext,
        ExperimentArms,
        FunnelSimple,
        PlaceboDescriptionInstruct,
        PlaceboDialogs,
        PlaceboKeywordBaseline,
        PlaceboSimpleInstruct,
        PlaceboTree,
        RankParagraphs,
    ]



from collections.abc import Sequence

from fvalues import F

from ice.paper import Paper
from ice.paper import Paragraph
from ice.recipe import recipe
from ice.recipes.meta.eval_paper_qa.qa_baseline import answer
from ice.utils import map_async


def make_classification_prompt(paragraph: Paragraph, question: str) -> str:
    return F(
        f"""Here is a paragraph from a research paper: "{paragraph}"

Question: Does this paragraph answer the question '{question}'? Say Yes or No.
Answer:"""
    )


async def classify_paragraph(paragraph: Paragraph, question: str) -> float:
    choice_probs, _ = await recipe.agent().classify(
        prompt=make_classification_prompt(paragraph, question),
        choices=(" Yes", " No"),
    )
    return choice_probs.get(" Yes", 0.0)


async def get_relevant_paragraphs(
    paper: Paper, question: str, top_n: int = 3
) -> list[Paragraph]:
    probs = await map_async(
        paper.paragraphs, lambda par: classify_paragraph(par, question)
    )
    sorted_pairs = sorted(
        zip(paper.paragraphs, probs), key=lambda x: x[1], reverse=True
    )
    return [par for par, prob in sorted_pairs[:top_n]]


async def answer_for_paper(
    paper: Paper, question: str = "What was the study population?", top_n: int = 3
) -> tuple[str, Sequence[str]]:
    relevant_paragraphs = await get_relevant_paragraphs(paper, question, top_n=top_n)
    relevant_str = F("\n\n").join(str(p) for p in relevant_paragraphs)
    response = await answer(context=relevant_str, question=question)
    return response, [str(p) for p in relevant_paragraphs]



import hashlib
import json
import threading

from abc import ABCMeta
from asyncio import create_task
from collections.abc import Callable
from contextvars import ContextVar
from functools import lru_cache
from functools import partial
from functools import wraps
from inspect import getdoc
from inspect import getsource
from inspect import isclass
from inspect import iscoroutinefunction
from inspect import isfunction
from inspect import Parameter
from inspect import signature
from time import monotonic_ns
from typing import Any
from typing import cast
from typing import IO
from typing import Optional
from typing import Union

from structlog import get_logger

from ice.json_value import JSONValue
from ice.json_value import to_json_value
from ice.server import ensure_server_running
from ice.server import is_server_running
from ice.settings import OUGHT_ICE_DIR
from ice.settings import server_url
from ice.settings import settings
from ice.utils import make_id

from .logging import log_lock

log = get_logger()


parent_id_var: ContextVar[str] = ContextVar("id")

traces_dir = OUGHT_ICE_DIR / "traces"
traces_dir.mkdir(parents=True, exist_ok=True)


class Trace:
    """
    Manages storing trace data to disk.
    All data is stored under self.dir.
    The primary metadata is stored in self.file, named trace.jsonl.
    Potentially large values which will be lazily loaded are stored in block_*.jsonl.
    Each block file is appended to until its size exceeds BLOCK_LENGTH.
    """

    BLOCK_LENGTH = 1024**2

    def __init__(self):
        self.id = make_id()
        self.dir = traces_dir / self.id
        self.dir.mkdir()
        self.file = self._open("trace")
        self.block_number = -1  # so that it starts at _open_block below
        self._open_block()
        self._lock = threading.Lock()

        # Keep an in-memory cache of recent block values (JSON strings)
        # keyed by their SHA256 hash, so equal values get the same address.
        # Apply lru_cache to the bound method to skip the self argument,
        # so instances of this class are not stored in the cache
        # and they can be garbage collected.
        self._write_block_value_cached = lru_cache(maxsize=1024)(
            self._write_block_value
        )
        # _write_block_value only takes one argument _string_hash,
        # so the potentially large values themselves don't live in memory.
        # The actual value is kept briefly in this attribute.
        self._current_block_value: str

        parent_id_var.set(self.id)
        log.info(f"Trace: {self.url}")
        threading.Thread(target=self._server_and_browser).start()

    @property
    def url(self) -> str:
        return f"{server_url()}/traces/{self.id}"

    def _server_and_browser(self):
        # We use this lock to prevent logging from here (which runs in a
        # background thread) from burying the input prompt in
        # [Settings.__get_and_store].
        with log_lock:
            is_running = None
            if settings.OUGHT_ICE_AUTO_SERVER:
                ensure_server_running()
                is_running = True

            if not settings.OUGHT_ICE_AUTO_BROWSER:
                return

            is_running = is_running or is_server_running()
            if not is_running:
                return

            import webbrowser

            log.info(
                "Opening trace in browser, set OUGHT_ICE_AUTO_BROWSER=0 to disable."
            )
            webbrowser.open(self.url)

    def _open(self, name: str) -> IO[str]:
        return open(self.dir / f"{name}.jsonl", "a")

    def _open_block(self):
        self.block_number += 1
        self.block_file = self._open(f"block_{self.block_number}")
        self.block_length = 0
        self.block_lineno = 0

    def add_to_block(self, value) -> tuple[int, int]:
        """
        Write the value x to the current block file as a single JSON line.
        """
        string = _encode_json(value)
        string_hash = hashlib.sha256(string.encode("utf8")).digest()
        with self._lock:
            self._current_block_value = string
            return self._write_block_value_cached(string_hash)

    def _write_block_value(self, _string_hash: bytes) -> tuple[int, int]:
        address = (self.block_number, self.block_lineno)
        s = self._current_block_value
        self.block_file.write(s)
        self.block_length += len(s)
        if self.block_length > self.BLOCK_LENGTH:
            self.block_file.write("end\n")
            self.block_file.close()
            self._open_block()
        else:
            self.block_file.flush()
            self.block_lineno += 1
        return address


trace_var: ContextVar[Optional[Trace]] = ContextVar("trace", default=None)


def enable_trace():
    trace_var.set(Trace())


def trace_enabled():
    return trace_var.get() is not None


def emit(value):
    if trc := trace_var.get():
        trc.file.write(_encode_json(value))
        trc.file.flush()


def emit_block(x) -> tuple[int, int]:
    if trc := trace_var.get():
        return trc.add_to_block(x)
    else:
        return 0, 0


Scalar = Union[bool, int, float, str, None]


def add_fields(**fields: Scalar):
    if trace_enabled():
        id = parent_id_var.get()
        emit({f"{id}.fields.{key}": value for key, value in fields.items()})


def _encode_json(x) -> str:
    # Note that sort_keys=True here could improve caching,
    # but it might make the output less readable.
    return json.dumps(to_json_value(x), separators=(",", ":")) + "\n"


# To add records to the trace, use the following code:
#
# async def f(self, arg1, arg2, record = recorder):
#     record(k1=v1, k2=v2)
#
# You can name 'record' anything and pass it any JSON-serializable keyword arguments.
#
# You MUST set its default value to 'recorder'. This allows @trace to know where to
# inject the recorder, and it ensures that your function will work even if tracing is
# disabled.


Recorder = Callable[..., None]

recorder: Recorder = lambda **kwargs: None


class _Recorder:
    def __init__(self, id: str):
        self.id = id

    def __call__(self, **kwargs):
        emit({f"{self.id}.records.{make_id()}": emit_block(kwargs)})

    def __repr__(self):
        # So this can be used in `diskcache()` functions
        # The diskcache implementation should probably not
        # depend on calling __repr__ but that
        # seems like a larger refactor
        return "ice.trace._Recorder"


def trace(fn):
    if isclass(fn):
        for key, value in fn.__dict__.items():
            if isfunction(value):
                setattr(fn, key, trace(value))
        return fn

    if not iscoroutinefunction(fn):
        return fn

    @wraps(fn)
    async def wrapper(*args, **kwargs):
        if not trace_enabled():
            return await fn(*args, **kwargs)

        @wraps(fn)
        async def inner_wrapper(*args, **kwargs):
            id = make_id()
            parent_id = parent_id_var.get()
            parent_id_var.set(id)

            sig = signature(fn)
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()
            arg_dict = {}
            recorder_name = None
            for k, v in bound_args.arguments.items():
                p = sig.parameters[k]
                if p.default is recorder:
                    recorder_name = k
                if p.kind == Parameter.VAR_KEYWORD:
                    arg_dict.update(v)
                else:
                    arg_dict[k] = v

            arg_dict_json = to_json_value(arg_dict)
            call_event = dict(
                parent=parent_id,
                start=monotonic_ns(),
                name=fn.__name__ if hasattr(fn, "__name__") else repr(fn),
                shortArgs=get_strings(arg_dict_json),
                func=emit_block(func_info(fn)),
                args=emit_block(arg_dict_json),
            )
            self = arg_dict.get("self")
            if self:
                call_event["cls"] = self.__class__.__name__

            emit(
                {
                    id: call_event,
                    f"{parent_id}.children.{id}": True,
                }
            )

            if recorder_name:
                kwargs[recorder_name] = _Recorder(id)

            result = await fn(*args, **kwargs)
            result_json = to_json_value(result)
            emit(
                {
                    f"{id}.result": emit_block(result_json),
                    f"{id}.shortResult": get_strings(result_json),
                    f"{id}.end": monotonic_ns(),
                }
            )
            return result

        return await create_task(inner_wrapper(*args, **kwargs))

    return wrapper


def to_json_serializable(self):
    namespace = dict(class_name=self.__class__.__name__)
    for attr in dir(self):
        if not attr.startswith("_"):
            value = getattr(self, attr)
            if isinstance(value, (bool, int, float, str)):
                namespace[attr] = value

    return namespace


class TracedABCMeta(ABCMeta):
    def __new__(mcls, name, bases, namespace):
        return super().__new__(
            mcls,
            name,
            bases,
            {k: trace(v) for k, v in namespace.items()}
            | dict(dict=to_json_serializable),
        )


class TracedABC(metaclass=TracedABCMeta):
    ...


# TODO this and the functions it calls needs to be replaced with a better system
#   for summarising args and return values
def get_strings(value: JSONValue) -> list[str]:
    """
    Represent the given value as a short list of short strings
    that can be stored directly in the central trace file and loaded eagerly in the UI.
    """
    if isinstance(value, dict) and "value" in value:
        value = value["value"]

    if isinstance(value, dict):
        value = {k: v for k, v in value.items() if k not in ("self", "record")}

    result = _get_first_descendant(value)

    if isinstance(result, tuple):
        result = list(result)
    if not (isinstance(result, list) and result):
        # if result in (None, (), "", [], {}):
        # but without breaking due to truth-testing of pandas dataframes
        if any(
            isinstance(result, type(x)) and result == x
            for x in cast(tuple[Any], (None, (), "", [], {}))
        ):
            result = "()"
        result = [str(result)]

    result = _get_short_list(result)
    result = [_get_short_string(v) for v in result]
    return result


def _get_short_string(string, max_length=35) -> str:
    return string[:max_length].strip() + "..." if len(string) > max_length else string


def _get_short_list(lst: list, max_length=3) -> list:
    return lst[:max_length] + ["..."] if len(lst) > max_length else lst


def _get_first_descendant(value: JSONValue) -> Any:
    if isinstance(value, dict) and value:
        first, *_ = value.values()
        return _get_first_descendant(first)
    elif isinstance(value, list) and value:
        if isinstance(value[0], str):
            return [v for v in value if isinstance(v, str)]
        return _get_first_descendant(value[0])
    else:
        return value


@lru_cache()
def func_info(fn):
    return dict(
        doc=getdoc(fn),
        source=getsource(fn.func) if isinstance(fn, partial) else getsource(fn),
    )



from dataclasses import dataclass
from typing import Protocol

from ice.recipes.primer.answer_by_computation import answer_by_computation
from ice.recipes.primer.answer_by_reasoning import answer_by_reasoning
from ice.recipes.primer.answer_by_search import answer_by_search


class QuestionRecipe(Protocol):
    async def __call__(self, question: str) -> str:
        ...


@dataclass
class Action:
    name: str
    description: str
    recipe: QuestionRecipe


action_types = [
    Action(
        name="Web search",
        description="Run a web search using Google. This is helpful if the question depends on obscure facts or current information, such as the weather or the latest news.",
        recipe=answer_by_search,
    ),
    Action(
        name="Computation",
        description="Run a computation in Python. This is helpful if the question depends on calculation or other mechanical processes that you can specify in a short program.",
        recipe=answer_by_computation,
    ),
    Action(
        name="Reasoning",
        description="Write out the reasoning steps. This is helpful if the question involves logical deduction or evidence-based reasoning.",
        recipe=answer_by_reasoning,
    ),
]



from ice.formatter.transform.value import numbered_list
from ice.formatter.transform.value import ValueTransform


def test_numbered_list():
    value = ["Paragraph Uno.", "Paragraph Dos  "]
    xform = numbered_list(value, separator="\n\n")
    assert isinstance(xform, ValueTransform)
    assert xform.transform() == "1. Paragraph Uno.\n\n2. Paragraph Dos"



from collections.abc import Callable
from collections.abc import Sequence
from collections.abc import Sized
from typing import Optional
from typing import Union

from structlog.stdlib import get_logger

from ice.formatter.multi import format_multi
from ice.formatter.multi import stop
from ice.formatter.multi import StopSentinel
from ice.formatter.transform.dependent import CountWord
from ice.formatter.transform.dependent import DependentTransform
from ice.formatter.transform.dependent import plural_transform
from ice.formatter.transform.positional import OrdinalWord
from ice.formatter.transform.value import numbered_list
from ice.formatter.transform.value import ValueTransform
from ice.recipes.experiments_and_arms.num_utils import extract_nums
from ice.recipes.experiments_and_arms.prompts.utils import get_part
from ice.recipes.experiments_and_arms.types import MultipartReasoningPrompt

log = get_logger()

# TODO: WRITE PROMPTS ##

NAME_EXPERIMENTS_PREFACE = """Some research papers cover just one experiment, while others cover multiple experiments. Experiments are distinct from trial arms or groups; a single experiment might have multiple trial arms, like different interventions or controls. Let's look at some excerpts from research papers covering different numbers of experiments and try to identify what those experiments were, being careful to not confuse experiments with trial arms.""".strip()

NAME_EXPERIMENTS_EXAMPLE_TEMPLATE = """
{ordinal}, consider the following excerpt{maybe_plural_excerpts} from a paper that conducted {experiments_count_word_or_just_one} experiment{maybe_plural_experiments}:

{paragraphs}

Let's think about what each excerpt tells us, if anything, about what {that_or_those} experiment{maybe_plural_experiments} {was_or_were}: {reasoning}

In summary, the {experiments_count_word} experiment{maybe_plural_experiments} {was_or_were}:

{answer}
""".strip()


NAME_EXPERIMENTS_EXAMPLES: list[
    dict[str, Union[ValueTransform[Sequence[str]], str, int]]
] = [
    dict(
        paragraphs=numbered_list(
            [
                "Contrary to expectations, participants in the park walking group reported an increase in fatigue at the end of the afternoon (d = -0.22) and the relaxation group experienced lower restoration in the evening compared to baseline (d = -0.18). In the evening the relaxation group experienced lower levels of fatigue (d = 0.25) than before the intervention. Across all time points of the working day, recovery experiences and outcomes in the week after the intervention were fairly similar to levels before the intervention in the three groups with average effect sizes around zero.",
                "After the intervention period the park walking group reported slightly lower levels of enjoyment of their lunch breaks (d = -0.38). The relaxation group experienced somewhat higher levels of detachment (d = 0.20) than before the intervention period. The control group's recovery experiences during lunch breaks were the same as before (d = -0.04). Concerning well-being outcomes, only the park walking group reported a positive change in restoration (d = 0.23) after lunch compared to baseline. All other changes after the lunch break were trivial in terms of effect sizes. In afternoons and evenings there were only three meaningful changes regarding well-being.",
                "The research plan for the RCT´s has been approved by the Ethical Committee of the Tampere Region, Finland. Data collection was divided into two phases to optimize our scarce material and personnel resources. The first RCT took place in spring (starting in week 18) and the second identical RCT in fall 2014 (starting in week 35). Thus, the second RCT in fall was a replication of the RCT in spring. Each RCT lasted five working weeks, two of which were the intervention period.",
                "There were three significant interaction terms of group x time, namely for detachment and enjoyment of the lunch break and for fatigue at the end of the afternoon (Table III ). The interaction effect for relaxation was marginally significant. The nature of these interactions will become clear in the following detailed description of the changes across time within the three groups.",
            ]
        ),
        reasoning="""Excerpt 1 discusses three groups, but does not distinguish between different experiments. Excerpt 2 also discusses two groups, a park walking group and a relaxation group, but the discussion suggests they belong to the same experiment. Excerpt 2 discusses two RCTs, one that took place in spring, and a second, which took place in fall 2014. These RCTs are separate experiments. Excerpt 4 discusses the three groups again but does not provide more information about the number of experiments.""",
        answer=numbered_list(["a spring RCT", "a fall 2014 RCT"]),
        experiments_count=2,
    ),
    dict(
        paragraphs=numbered_list(
            [
                "Patients were evaluated at baseline, 8 hours after the first administration of the study drug (tolvaptan or placebo), and on days 2, 3, 4, 11, 18, 25, 30, and 37 . Study drugs were withheld after day 30, and the effect of discontinuation of the study drug was assessed on day 37.",
                "The present study was conducted primarily in the outpatient setting, without mandated fluid restriction or a change in the patient's medication regimen, such as use of diuretics, to treat the patient's primary disease. Tolvaptan was superior to placebo with respect to several measures, including the change in the average daily AUC for serum sodium concentrations from baseline to day 4 and from baseline to day 30, the mean serum sodium concentration at each visit, the time to normalized serum sodium concentrations, the percentage of patients with serum sodium concentrations that were normal on day 4 and on day 30, and the categorical change in the serum sodium concentration from baseline to day 4 and from baseline to day 30. Tolvaptan was superior to placebo from the first observation point (8 hours) after administration of the first dose until the last treatment day (day 30) in patients with either mild or marked hyponatremia and among patients with hyponatremia from all major causes. During the 7-day follow-up period, serum sodium concentrations reverted to degrees of hyponatremia that were equivalent to those associated with the use of placebo, indicating that the aquaretic effect of tolvaptan (excretion of electrolyte-free water) was required to maintain normal sodium concentrations in patients with chronic hyponatremia.",
                "Patients were ineligible if they had clinically evident hypovolemic hyponatremia (a state in which normal plasma sodium concentrations could be reestablished through the restoration of plasma volume).",
                "Hyponatremia occurs in 15 to 20% of hospitalized patients and constitutes a common serum electrolyte abnormality. 22  Hyponatremia is reported to be an independent predictor of complications and death in patients with heart disease, 23,24 cirrhosis, 6 or neurologic disorders.",
            ]
        ),
        reasoning="""Excerpt 1 discusses two trial arms (tolvaptan and placebo). Because this paper studied one experiment, we can surmise that it was a study of tolvaptan. Excerpt 2 more information about the study. Excerpt 3 discusses exclusion criteria for participants. Expert 4 provides information about hyponatremia.""",
        answer=numbered_list(["a tolvaptan study"]),
        experiments_count=1,
    ),
]

NAME_EXPERIMENTS_SHARED: dict[
    str, Union[OrdinalWord, DependentTransform[Union[int, Sized]]]
] = dict(
    ordinal=OrdinalWord(capitalize=True, finally_case="Finally"),
    maybe_plural_excerpts=plural_transform(
        key="paragraphs", singular_case="", plural_case="s"
    ),
    experiments_count_word_or_just_one=CountWord("experiments_count", {1: "just one"}),
    maybe_plural_experiments=plural_transform(
        key="experiments_count", singular_case="", plural_case="s"
    ),
    that_or_those=plural_transform(
        key="experiments_count", singular_case="that", plural_case="those"
    ),
    was_or_were=plural_transform(
        key="experiments_count", singular_case="was", plural_case="were"
    ),
    experiments_count_word=CountWord(key="experiments_count"),
)


def count_from_answer(answer: str) -> int:
    nums = extract_nums(answer)
    if not nums:
        log.warning("No number in final answer; expected count", answer=answer)
    return nums[-1] if nums else 0


NAME_EXPERIMENTS_REASONING_STOP = ("\n\nIn summary",)


def make_name_exps_from_count(
    experiments_count: int,
) -> Callable[[int], MultipartReasoningPrompt]:
    def make_name_experiments_prompt_func(num_shots: int) -> MultipartReasoningPrompt:
        def name_experiments_prompt(
            paragraphs: Sequence[str],
            helpfulness: Optional[str] = None,
            reasoning: Optional[str] = None,
        ) -> str:
            helpfulness  # ignored
            last_example: dict[
                str, Union[ValueTransform[Sequence[str]], str, StopSentinel, int]
            ] = dict(
                paragraphs=numbered_list(paragraphs),
                reasoning=reasoning if reasoning else stop("Excerpt 1"),
                experiments_count=experiments_count,
            )
            if reasoning:
                last_example["answer"] = stop("")
            shots = format_multi(
                NAME_EXPERIMENTS_EXAMPLE_TEMPLATE,
                NAME_EXPERIMENTS_EXAMPLES[:num_shots] + [last_example],
                NAME_EXPERIMENTS_SHARED,
            )
            return "\n\n".join([NAME_EXPERIMENTS_PREFACE] + list(shots))

        return name_experiments_prompt

    return make_name_experiments_prompt_func


def get_name_exps_reasoning(response: str) -> str:
    return "".join(("Excerpt 1", get_part(response, ": ", "\nIn summary")))



from collections.abc import Sequence
from typing import Generic
from typing import Optional

from pydantic.generics import GenericModel

from ice.recipes.program_search.utils.find_examples import mark_gs
from ice.recipes.program_search.utils.find_examples import rouge_distractor_scores
from ice.recipes.program_search.utils.find_examples import SelectionT_co
from ice.utils import window_dropping


class SelectionExample(GenericModel, Generic[SelectionT_co]):
    selection: Sequence[SelectionT_co]
    existing: Sequence[str]
    positive_idxs: Sequence[int]
    distractor_score: float


async def make_examples(
    texts: Sequence[SelectionT_co],
    gs_quotes: Sequence[str],
    *,
    n: int,
    step: int,
    max_existing: int,
) -> Sequence[SelectionExample]:
    gs_labeled = await mark_gs(texts, gs_quotes)
    gs_texts = {str(text) for text in gs_labeled if text.is_gs}
    windowed = window_dropping(gs_labeled, n, step)
    examples: list[SelectionExample] = []
    for window in windowed:
        # TODO: Consider different orderings?
        positive_idxs = [idx for idx, text in enumerate(window) if text.is_gs]
        existing = list(gs_texts - {str(text) for text in window})[:max_existing]
        distractor_scores = await rouge_distractor_scores(
            [text.original for text in window], references=gs_quotes
        )
        summary_distractor_score = (
            sum(distractor_scores.values()) / len(distractor_scores)
            if distractor_scores
            and not any(score == 0 for score in distractor_scores.values())
            else 0
        )
        examples.append(
            SelectionExample(
                selection=window,
                existing=existing,
                positive_idxs=positive_idxs,
                distractor_score=summary_distractor_score,
            )
        )
    return examples


def first_positive_example(
    examples: Sequence[SelectionExample],
) -> Optional[SelectionExample]:
    try:
        return next(
            filter(
                lambda example: example.positive_idxs,
                examples,
            )
        )
    except StopIteration:
        return None


def best_negative_example(
    examples: Sequence[SelectionExample],
) -> Optional[SelectionExample]:
    negative_examples = list(
        filter(lambda example: not example.positive_idxs, examples)
    )
    if not negative_examples:
        return None
    return max(negative_examples, key=lambda example: example.distractor_score)



from collections.abc import Sequence

from ice.formatter.multi import format_multi
from ice.formatter.multi import stop
from ice.formatter.transform.value import numbered_list

INSTRUCTIONS = """For each pair of lists, can each item in list A be paired with a corresponding item in list B that is substantially similar to it? If so, explain why. If not, explain why not."""

EXAMPLE_TEMPLATE = """
List A:

{ground_truth_items}

List B:

{predicted_items}

Let's think it over. {reasoning}

Final answer: {final_answer}
""".strip()

YES_ANSWER = (
    "Yes, each item in list A can be paired with a corresponding item in list B."
)
NO_ANSWER = "No, it is not the case that each item in list A can be paired with a corresponding item in list B."

EXAMPLES = [
    dict(
        ground_truth_items=numbered_list(
            [
                "SALT-1: multicenter, randomized, double-blind, placebo-controlled trial of tolvaptan for hyponatremia",
                "SALT-2: multicenter, randomized, double-blind, placebo-controlled trial of tolvaptan for hyponatremia",
            ]
        ),
        predicted_items=numbered_list(
            [
                "a study of tolvaptan",
                "a study of the effect of discontinuation of the study drug",
            ]
        ),
        reasoning="""The items in List A only differ by SALT-1 versus SALT-2, but the items in List B do not share this difference. Therefore, we cannot match the items in List A up with the items in List B.""",
        final_answer=NO_ANSWER,
    ),
    dict(
        ground_truth_items=numbered_list(
            [
                "preintervention cases: a stepped wedge cluster RCT to examine the effect of the SSC on in-hospital complications",
                "postintervention cases: a stepped wedge cluster RCT to examine the effect of the SSC on in-hospital complications",
            ]
        ),
        predicted_items=numbered_list(
            [
                "study of the SSC before surgery",
                "study of the SSC after surgery",
            ]
        ),
        reasoning="""The items in List A differ by mentioning preintervention and postintervention, and they are both about studies of the effect of the SSC. \
The items in List B differ in being before (pre) or after (post) surgery, and are both about studies of the SSC. Therefore, we can match the \
"preintervention cases" item in List A to the "before surgery" item in List B, and the "postintervention cases" item in List A to the \
"after surgery" item in List B.""",
        final_answer=YES_ANSWER,
    ),
]


def matching_prompt(items_a: Sequence[str], items_b: Sequence[str]) -> str:
    examples = EXAMPLES + [
        dict(ground_truth_items=items_a, predicted_items=items_b, reasoning=stop(""))
    ]
    formatted_examples = format_multi(EXAMPLE_TEMPLATE, examples)  # type: ignore[arg-type]
    return "\n\n".join([INSTRUCTIONS, "\n\n---\n\n".join(formatted_examples)])


MATCHING_STOP_SEQUENCES = [
    ", the items in list A can be matched with the items in list B.",
    "the items in list A cannot be matched with the items in list B.",
]


def reasoning_and_answer_from_completion(completion: str) -> tuple[str, bool]:
    reasoning_part, answer_part = completion.split("Final answer: ")
    yes_answer = "Yes" in answer_part
    return reasoning_part.strip(), yes_answer



from typing import Optional

from structlog.stdlib import get_logger

from ice.recipe import Recipe
from ice.recipes.evaluate_result import EvaluateResult
from ice.recipes.evaluate_result import ResultComparison
from ice.utils import map_async

log = get_logger()


class EvaluateResults(Recipe):
    async def run(
        self,
        question: Optional[str] = None,
        model_results: Optional[list[str]] = None,
        gold_results: Optional[list[str]] = None,
    ) -> list[ResultComparison]:
        """
        Compare two lists of results, model and gold standard.
        """
        evaluate_result = EvaluateResult(mode=self.mode)

        if not model_results and not gold_results and not question:
            if not self.mode == "test":
                log.warning("No model results and no gold results - using test data.")
            model_results, gold_results, question = evaluate_result.test_data(n=3)
        elif not model_results or not gold_results or not question:
            raise ValueError("Must provide both model results and gold results.")

        comparisons = await map_async(
            list(zip(model_results, gold_results)),
            lambda pair: evaluate_result.run(
                question=question, model_result=pair[0], gold_result=pair[1]
            ),
            max_concurrency=self.max_concurrency(),
            show_progress_bar=True,
        )

        return comparisons



from fvalues import F

from ice.recipe import recipe


def generate_reasoning_prompt(question: str) -> str:
    return F(
        f"""Answer the following question:

Question: "{question}"
Answer: "Let's think step by step.
"""
    ).strip()


def generate_answer_prompt(question: str, reasoning: str) -> str:
    return F(
        f"""Answer the following question using the reasoning shown below:

Question: "{question}"
Reasoning: "{reasoning}"
Short answer: "
"""
    ).strip()


async def get_reasoning(question: str) -> str:
    reasoning_prompt = generate_reasoning_prompt(question)
    reasoning = await recipe.agent().complete(prompt=reasoning_prompt, stop='"')
    return reasoning


async def get_answer(question: str, reasoning: str) -> str:
    answer_prompt = generate_answer_prompt(question, reasoning)
    answer = await recipe.agent().complete(prompt=answer_prompt, stop='"')
    return answer


async def answer_by_reasoning(
    question: str = "What would happen if the average temperature in Northern California went up by 5 degrees Fahrenheit?",
) -> str:
    reasoning = await get_reasoning(question)
    answer = await get_answer(question, reasoning)
    return answer


recipe.main(answer_by_reasoning)



from fvalues import F

DEFAULT_QUESTION = "Beth bakes 4x 2 dozen batches of cookies in a week. If these cookies are shared amongst 16 people equally, how many cookies does each person consume?"

DEFAULT_STEPS = [
    "Beth bakes 4x 2 dozen batches of cookies for a total of 4*2 = 8 dozen cookies",
    "There are 12 cookies in a dozen and she makes 8 dozen cookies for a total of 12*8 = 96 cookies",
    "She splits the 96 cookies equally amongst 16 people so they each eat 96/16 = 6 cookies",
    "So, the final answer is 6 cookies per person.",
]


def render_steps(steps: list[str]) -> str:
    return F("\n").join(F(f"{i}. {step}") for (i, step) in enumerate(steps, start=1))



from collections.abc import Sequence

from structlog.stdlib import get_logger


log = get_logger()


CONSENSUS_PREFIX = "Below are some answers to the question '{question}'"
CLUSTER_DIRECTIONS = """Cluster these answers by how they answer the question '{question}'. Write your answer in the form "Cluster N) [Answer]":

Cluster 1)"""

FINAL_DIRECTIONS = """Based on these clusters, write a single, detailed answer based on a consensus of the proposed answers, that is most likely to be correct.

Detailed, consensus answer:"""


def candidates(
    answers: Sequence[str], prefix: str = "Answer", separator: str = "\n"
) -> str:
    return separator.join(
        (f"{prefix} {idx}) {answer.strip()}" for idx, answer in enumerate(answers, 1))
    )


def build_cluster_prompt(question: str, answers: Sequence[str]) -> str:
    prefix = CONSENSUS_PREFIX.format(question=question)
    return "\n\n".join(
        (prefix, candidates(answers), CLUSTER_DIRECTIONS.format(question=question))
    )


def build_final_prompt(question: str, answers: Sequence[str], clusters: str) -> str:
    cluster_prompt = build_cluster_prompt(question, answers)
    return "\n\n".join((cluster_prompt + clusters.rstrip(), FINAL_DIRECTIONS))
